from __future__ import annotations

from contextlib import nullcontext
from pathlib import Path
from typing import Dict, List, Optional, Tuple, Union

import rich.console
import rich.layout
import rich.live
import rich.panel
import rich.tree
from PIL import Image
from typing_extensions import Self

from verdict.core.executor import GraphExecutor
from verdict.core.primitive import Block, Layer, MaterializationContext, Unit
from verdict.core.visualization import StreamingLayoutManager
from verdict.dataset import DatasetWrapper
from verdict.model import ModelSelectionPolicy
from verdict.schema import Schema
from verdict.util.exceptions import VerdictDeclarationTimeError
from verdict.util.log import init_logger, logger
from verdict.util.misc import keyboard_interrupt_safe
from verdict.util.tracing import (
    ExecutionContext,
    Tracer,
    TracingManager,
    ensure_tracing_manager,
)


class Pipeline:
    name: str
    block: Block
    executor: GraphExecutor
    default_tracer: Optional[Union[Tracer, List[Tracer]]]

    def __init__(
        self,
        name: str = "Pipeline",
        tracer: Optional[Union[Tracer, List[Tracer]]] = None,
    ) -> None:
        """
        Initialize a Pipeline.

        Args:
            name: The name of the pipeline.
            tracer: The default tracer or list of tracers to use. If None, uses NoOpTracer().
        """
        super().__init__()
        self.name = name
        self.block = Block()
        self.default_tracer: TracingManager = ensure_tracing_manager(tracer)

    def add_tracer(self, tracer: Tracer) -> None:
        """
        Add a tracer to the pipeline's TracingManager.

        Args:
            tracer: The tracer to add.
        """
        self.default_tracer.add_tracer(tracer)

    def copy(self) -> "Pipeline":
        pipeline = Pipeline(self.name, tracer=self.tracer)
        pipeline.block = self.block.copy()
        return pipeline

    def __rshift__(self, other: Union[Unit, Layer, "Block"]) -> Self:
        self.block >>= other
        return self

    def via(
        self,
        policy_or_name: Union[ModelSelectionPolicy, str],
        retries: int = 1,
        **inference_parameters,
    ) -> Self:
        self.block.via(policy_or_name, retries, **inference_parameters)
        return self

    def collect_outputs(
        self, executor: GraphExecutor, block_instance: Block
    ) -> Tuple[Dict[str, Schema], List[str]]:
        leaf_node_prefixes: List[str] = []
        outputs: Dict[str, Schema] = {}

        for node in block_instance.nodes:
            try:
                if node not in executor.outputs:
                    logger.debug(f"Node {node} has no output!")
                    continue

                if not isinstance(executor.outputs[node], Schema):  # type: ignore
                    logger.debug(
                        f"Node {node} with output {executor.outputs[node]} is not a Schema"
                    )  # type: ignore

                for key, value in executor.outputs[node].model_dump().items():  # type: ignore
                    outputs[
                        (column_name := f"{self.name}_{'.'.join(node.prefix)}_{key}")
                    ] = value  # type: ignore
                    if node in block_instance.leaf_nodes:
                        leaf_node_prefixes.append(column_name)
            except Exception as e:
                raise VerdictDeclarationTimeError(
                    f"Error collecting outputs for node {node} with outputs {executor.outputs[node]}"
                ) from e  # type: ignore

        return outputs, sorted(list(leaf_node_prefixes))

    @keyboard_interrupt_safe
    def run(
        self,
        input_data: Schema = Schema.empty(),
        max_workers: int = 128,
        display: bool = False,
        graceful: bool = False,
        tracers: Optional[Union[Tracer, List[Tracer]]] = None,
    ) -> Tuple[Dict[str, Schema], List[str]]:
        """
        Run the pipeline.

        Args:
            input_data: The input schema.
            max_workers: Number of workers.
            display: Whether to display live output.
            graceful: Whether to shut down gracefully.
            tracers: Optional tracer(s) to override the pipeline's tracers for this run.

        Returns:
            Tuple of outputs and leaf node prefixes.
        """
        self.block = self.block.copy()
        init_logger(self.name)
        logger.info(f"Starting pipeline {self.name}")
        if tracers is not None:
            self.default_tracer = ensure_tracing_manager(tracers)
        tracer_to_use = self.default_tracer
        execution_context = ExecutionContext(
            tracer=tracer_to_use,
            parent_id=None,
        )
        self.executor = GraphExecutor(
            max_workers=max_workers, execution_context=execution_context
        )

        with execution_context.trace_call(
            name=f"{getattr(self, 'name', None) or self.__class__.__name__}",
            inputs={
                "input_data": input_data,
                "max_workers": max_workers,
                "display": display,
                "graceful": graceful,
            },
        ) as call:
            with (
                rich.live.Live(auto_refresh=True) if display else nullcontext()
            ) as live:  # type: ignore
                context = MaterializationContext(n_instances=1)
                if display:
                    context.parent_branch = rich.tree.Tree(self.name)
                    context.streaming_layout_manager = StreamingLayoutManager(
                        rich.layout.Layout()
                    )

                    layout = rich.layout.Layout()
                    layout.split_row(
                        rich.layout.Layout(
                            rich.panel.Panel(
                                context.parent_branch, title="Execution Tree"
                            ),
                            name="execution",
                        ),
                        rich.layout.Layout(
                            rich.panel.Panel(
                                context.streaming_layout_manager.layout,
                                title="Streaming",
                            ),
                            name="streaming",
                        ),
                    )

                    live.update(layout)  # type: ignore

                block_instance, _ = self.block._materialize(context.copy())
                block_instance.source_input = input_data
                block_instance.executor = self.executor

                self.executor.submit(
                    block_instance.root_nodes,
                    input_data,
                    trace_id=execution_context.trace_id,
                    parent_id=call.call_id
                    if call is not None
                    else execution_context.call_id,
                )  # type: ignore
                self.executor.wait_for_completion(graceful=graceful)

            logger.info(f"Pipeline {self.name} completed")
            outputs = self.collect_outputs(self.executor, block_instance)
            if call is not None:
                call.set_outputs(outputs)
            return outputs

    @keyboard_interrupt_safe
    def run_from_dataset(
        self,
        dataset: DatasetWrapper,
        max_workers: int = 128,
        experiment_config=None,
        display: bool = False,
        graceful: bool = False,
        tracers: Optional[Union[Tracer, List[Tracer]]] = None,
    ) -> Tuple["pd.DataFrame", List[str]]:
        """
        Run the pipeline on a dataset.

        Args:
            dataset: The dataset wrapper.
            max_workers: Number of workers.
            experiment_config: Experiment config.
            display: Whether to display live output.
            graceful: Whether to shut down gracefully.
            tracers: Optional tracer(s) to override the pipeline's tracers for this run.

        Returns:
            Tuple of DataFrame and leaf node prefixes.
        """
        self.block = self.block.copy()
        init_logger(self.name)
        logger.info(f"Running pipeline {self.name} on dataset (len={len(dataset)})")
        if tracers is not None:
            self.default_tracer = ensure_tracing_manager(tracers)
        tracer_to_use = self.default_tracer
        execution_context = ExecutionContext(
            tracer=tracer_to_use,
            parent_id=None,
        )
        self.executor = GraphExecutor(
            max_workers=max_workers, execution_context=execution_context
        )

        dataset_df = dataset.samples.copy()
        with execution_context.trace_call(
            name=f"{getattr(self, 'name', None) or self.__class__.__name__}",
            inputs={
                "dataset": dataset_df,
                "max_workers": max_workers,
                "experiment_config": experiment_config,
                "display": display,
                "graceful": graceful,
            },
        ) as call:
            with (
                rich.live.Live(auto_refresh=True) if display else nullcontext()
            ) as live:  # type: ignore
                block_instances: Dict[str, Block] = {}

                context = MaterializationContext(n_instances=len(dataset))
                if display:
                    context.parent_branch = rich.tree.Tree(self.name)
                    context.streaming_layout_manager = StreamingLayoutManager(
                        rich.layout.Layout()
                    )

                    execution_layout = rich.layout.Layout(name="execution")
                    execution_tree_layout = rich.layout.Layout(
                        rich.panel.Panel(context.parent_branch, title="Execution Tree"),
                        ratio=1,
                        name="execution",
                    )
                    if experiment_config:
                        from verdict.util.experiment import get_experiment_layout

                        execution_layout.split_column(
                            get_experiment_layout(dataset_df, experiment_config),
                            execution_tree_layout,
                        )
                    else:
                        execution_layout.add_split(execution_tree_layout)

                    layout = rich.layout.Layout()
                    layout.split_row(
                        execution_layout,
                        rich.layout.Layout(
                            rich.panel.Panel(
                                context.streaming_layout_manager.layout,
                                title="Streaming",
                            ),
                            name="streaming",
                        ),
                    )

                    live.update(layout)  # type: ignore

                prototype, _ = self.block._materialize(context)

                for idx, (row, input_data) in enumerate(dataset):
                    block_instance = block_instances[row["hash(row)"]] = (
                        prototype.clone()
                    )
                    block_instance.source_input = input_data
                    block_instance.executor = self.executor

                    self.executor.submit(
                        block_instance.root_nodes,
                        input_data,
                        trace_id=execution_context.trace_id,
                        parent_id=call.call_id
                        if call is not None
                        else execution_context.call_id,
                    )  # type: ignore

                self.executor.wait_for_completion(graceful=graceful)

                output = []
                leaf_node_prefixes = set()
                for row_id, block_instance in block_instances.items():
                    row_output = {"hash(row)": row_id}
                    outputs, _leaf_node_prefixes = self.collect_outputs(
                        self.executor, block_instance
                    )

                    row_output.update(outputs)
                    leaf_node_prefixes.update(_leaf_node_prefixes)

                    if len(row_output) > 1:
                        output.append(row_output)

                if len(output) > 0:
                    import pandas as pd

                    result_df = pd.merge(
                        dataset_df, pd.DataFrame(output), on="hash(row)", how="right"
                    ).drop(columns=["hash(row)"])
                else:
                    result_df = dataset_df

                if display and experiment_config:
                    from verdict.util.experiment import get_experiment_layout

                    execution_layout["experiment"].update(
                        get_experiment_layout(result_df, experiment_config)
                    )
                    live.refresh()  # type: ignore

                return result_df, sorted(list(leaf_node_prefixes))

    @keyboard_interrupt_safe
    def run_from_list(
        self,
        dataset: List[Schema],
        max_workers: int = 128,
        experiment_config=None,
        display: bool = False,
        graceful: bool = False,
        tracers: Optional[Union[Tracer, List[Tracer]]] = None,
    ) -> Tuple[Dict[str, Schema], List[str]]:
        """
        Run the pipeline on a list of Schemas.

        Args:
            dataset: List of input schemas.
            max_workers: Number of workers.
            experiment_config: Experiment config.
            display: Whether to display live output.
            graceful: Whether to shut down gracefully.
            tracers: Optional tracer(s) to override the pipeline's tracers for this run.

        Returns:
            Tuple of outputs and leaf node prefixes.
        """
        from datasets import Dataset  # type: ignore[import-untyped]

        vedict_dataset = DatasetWrapper(
            Dataset.from_list([data.model_dump() for data in dataset])
        )
        return self.run_from_dataset(
            vedict_dataset,
            max_workers,
            experiment_config,
            display,
            graceful,
            tracers=tracers,
        )

    def checkpoint(self, path: Path):
        # TODO
        pass

    def restore(self, path: Path):
        # TODO
        pass

    def plot(self, display=False) -> Image.Image:
        return self.block.materialize().plot(display)



import re
from abc import ABC, abstractmethod
from enum import Enum
from typing import Any, Dict, List, Optional, Tuple, Type, Union

from pydantic import Field
from pydantic.fields import FieldInfo

from verdict.util.exceptions import ConfigurationError


class ScaleType(Enum):
    DISCRETE = "discrete"
    CONTINUOUS = "continuous"


class Scale(ABC):
    T: Type

    def __init__(self, scale_type: ScaleType, end_is_worst: bool = False) -> None:
        self.scale_type = scale_type
        self.end_is_worst = end_is_worst
        self.signifiers = ["best", "worst"] if end_is_worst else ["worst", "best"]

    @abstractmethod
    def pydantic_fields(self, key: str = "output") -> Dict[str, Tuple[Any, Any]]:
        pass

    @abstractmethod
    def prompt(self) -> str:
        pass

    @abstractmethod
    def __str__(self) -> str:
        pass


class DiscreteScale(Scale):
    values: List[Any]

    def __init__(
        self,
        values: Union[List[Any], Tuple[Any, Any], Tuple[Any, Any, Optional[int]]],
        end_is_worst: bool = False,
    ):
        super().__init__(ScaleType.DISCRETE, end_is_worst)
        if isinstance(values, tuple):
            if len(values) == 2:
                values = [*values, 1]
            self.start, self.end, step = values
            if step is None:
                step = 1

            self.T = type(self.start)
            if self.T is int:
                self.values = list(range(self.start, self.end + 1, step))
            elif self.T is float:
                self.values = [
                    round(self.start + i * step, 10)
                    for i in range(int((self.end - self.start) / step) + 1)
                ]
            elif self.T is str:
                self.values = [
                    chr(i) for i in range(ord(self.start), ord(self.end) + 1, step)
                ]
            else:
                raise ConfigurationError(
                    "Invalid values for DiscreteScale: must be a list or a tuple defining a range"
                )
        elif isinstance(values, list):
            self.values = values
        else:
            raise ConfigurationError(
                "Invalid values for DiscreteScale: must be a list or a tuple defining a range"
            )

        if not self.values:
            raise ConfigurationError("Discrete scale must have at least one value")

        self.T = type(self.values[0])

    def pydantic_fields(self, key: str = "output") -> Dict[str, Tuple[Any, FieldInfo]]:
        if self.T in [int, float] and hasattr(self, "start") and hasattr(self, "end"):
            output_field = Field(..., ge=self.start, le=self.end)
        else:
            value_options = "|".join(re.escape(str(value)) for value in self.values)
            output_field = Field(..., pattern=f"^({value_options})$")

        return {
            key: (self.T, output_field),
        }

    def prompt(self) -> str:
        return f"""Respond with a single output value from "{'", "'.join(map(str, self.values))}" where "{self.values[0]}" is {self.signifiers[0]} and "{self.values[-1]}" is {self.signifiers[1]}."""

    def token_support(self) -> List[str]:
        return [str(value) for value in self.values]

    def index(self, token: str) -> int:
        return self.token_support().index(token)

    def value_mapping_fn(self, output: str) -> Any:
        if self.T is int:
            return int(output)

        if self.T is float:
            return float(output)

        return output

    def __str__(self) -> str:
        return ", ".join(
            map(str, self.values[::-1] if self.end_is_worst else self.values)
        )


class BooleanScale(DiscreteScale):
    def __init__(
        self, yes: List[str] = ["yes", "Yes", "YES"], no: List[str] = ["no", "No", "NO"]
    ) -> None:
        values = no + yes
        super().__init__(values=values)

        self.yes = yes
        self.no = no

        self.T = bool

    def pydantic_fields(self, key: str = "output") -> Dict[str, Tuple[Any, FieldInfo]]:
        return {key: (bool, Field(...))}

    def token_support(self) -> List[str]:
        return self.yes + self.no

    def value_mapping_fn(self, output: str) -> bool:
        if output in self.yes:
            return True
        return False

    def __str__(self) -> str:
        return ",".join(map(lambda v: "/".join(v), [self.yes, self.no]))


class ContinuousScale(Scale):
    def __init__(
        self, min_value: float, max_value: float, end_is_worst: bool = False
    ) -> None:
        super().__init__(ScaleType.CONTINUOUS, end_is_worst)
        if min_value >= max_value:
            raise ConfigurationError(
                "Continuous scale minimum must be less than maximum"
            )
        self.min_value = min_value
        self.max_value = max_value

        self.T = float

    def pydantic_fields(self, key: str = "output") -> Dict[str, Tuple[Any, FieldInfo]]:
        return {
            key: (float, Field(..., ge=self.min_value, le=self.max_value)),
        }

    def prompt(self) -> str:
        return f"Provide a score between {self.min_value} ({self.signifiers[0]}) and {self.max_value} ({self.signifiers[1]})."

    def value_mapping_fn(self, output: float) -> float:
        return output

    def __str__(self) -> str:
        return (
            f"{self.max_value} - {self.min_value}"
            if self.end_is_worst
            else f"{self.min_value} - {self.max_value}"
        )


def LikertScale(end_is_worst: bool = False) -> Scale:
    return DiscreteScale((1, 5), end_is_worst)



from __future__ import annotations

from pathlib import Path
from typing import Any, Callable, Dict, Iterator, List, Optional, Tuple

import dill  # type: ignore[import-untyped]
from typing_extensions import Self

from verdict.schema import Schema
from verdict.util.exceptions import ConfigurationError

InputFn = Callable[[Dict[str, Any]], Schema]


class DatasetWrapper(Iterator[Tuple[Dict[str, Any], Schema]]):
    dataset: "pd.DataFrame"  # type: ignore
    samples: "pd.DataFrame"  # type: ignore

    max_samples: Optional[int]

    input_fn: InputFn

    _iter: Optional[Iterator[Tuple[Dict[str, Any], Schema]]] = None
    _count: int = 0

    def __init__(
        self,
        dataset: "Dataset",  # type: ignore
        input_fn: Optional[InputFn] = None,
        columns: Optional[List[str]] = None,
        max_samples: Optional[int] = None,
    ):
        import pandas as pd

        self.dataset = pd.DataFrame(dataset)
        self.dataset["hash(row)"] = self.dataset.apply(
            lambda row: hash(str(row)), axis=1
        )
        self.max_samples = int(max_samples) if max_samples else None
        self.samples = (
            self.dataset.sample(n=self.max_samples)
            if self.max_samples
            else self.dataset
        )

        if input_fn is not None and columns is not None:
            raise ConfigurationError("Cannot specify both input_fn and columns")

        columns = columns or (self.dataset.columns.difference(["hash(row)"]).tolist())
        self.input_fn = input_fn or (
            lambda row: Schema.of(**{k: v for k, v in row.items() if k in columns})
        )

    def __iter__(self) -> Self:
        samples = []
        for idx, row in self.samples.iterrows():
            samples.append((_row := row.to_dict(), sample := self.input_fn(_row)))
            for key, value in sample.model_dump().items():
                column = f"!{key}"
                if column not in self.samples.columns:
                    self.samples[column] = None

                self.samples.at[idx, column] = value

        self._iter = iter(samples)
        self._count = 0
        return self

    def __next__(self) -> Tuple[Dict[str, Any], Schema]:
        if self._iter is None:
            raise StopIteration

        if self.max_samples and self._count >= self.max_samples:
            self._iter = None
            raise StopIteration

        self._count += 1
        return next(self._iter)

    def __len__(self) -> int:
        return len(self.samples)

    def save(self, path: Path) -> None:
        with open(path, "wb") as f:
            dill.dump(self, f)

    @staticmethod
    def load(path: Path) -> "DatasetWrapper":
        with open(path, "rb") as f:
            return dill.load(f)

    @staticmethod
    def from_hf(
        dataset: Dict[str, "Dataset"],  # type: ignore
        input_fn: Optional[InputFn] = None,
        columns: Optional[List[str]] = None,
        max_samples: Optional[int] = None,
        expand: bool = False,
    ) -> Dict[str, "DatasetWrapper"]:
        from datasets import Dataset  # type: ignore[import-untyped]

        def expand_dataset(dataset: Dataset) -> Dataset:
            expanded_data = []
            for row in dataset:
                max_len = max(
                    len(value) if isinstance(value, list) else 1
                    for value in row.values()
                )
                for i in range(max_len):
                    new_row = {
                        key: value[i]
                        if isinstance(value, list) and i < len(value)
                        else value
                        for key, value in row.items()
                    }

                    for key, value in new_row.items():
                        if isinstance(value, list):
                            new_row[key] = str(value)
                    expanded_data.append(new_row)

            expanded_dict = {
                key: [row[key] for row in expanded_data] for key in expanded_data[0]
            }
            return Dataset.from_dict(expanded_dict)

        return {
            k: DatasetWrapper(
                expand_dataset(v) if expand else v, input_fn, columns, max_samples
            )
            for k, v in dataset.items()
        }

    @staticmethod
    def from_pandas(
        df: "pd.DataFrame",  # type: ignore
        input_fn: Optional[InputFn] = None,
        columns: Optional[List[str]] = None,
        split_column: Optional[str] = None,
        max_samples: Optional[int] = None,
    ):
        from datasets import Dataset  # type: ignore[import-untyped]

        if split_column is not None:
            return DatasetWrapper.from_hf(
                {
                    str(split): Dataset.from_pandas(_df)
                    for split, _df in df.groupby(split_column)
                },
                input_fn,
                columns,
                max_samples,
            )

        return DatasetWrapper.from_hf(
            {"all": Dataset.from_pandas(df)}, input_fn, columns, max_samples
        )



from dataclasses import dataclass, field
from typing import List

from verdict import Block, Layer, Pipeline
from verdict.common.judge import CategoricalJudgeUnit
from verdict.scale import DiscreteScale
from verdict.schema import Schema
from verdict.transform import MaxPoolUnit
from verdict.util import ratelimit

ratelimit.disable()

# Default prompts
DEFAULT_JUDGE_PROMPT = """
Determine whether the provided claim is consistent with the corresponding document. Consistency in this context implies that
all information presented in the claim is substantiated by the document. If not, it should be considered inconsistent.
Document: {source.doc}
Claim: {source.claim}
Please assess the claim's consistency with the document by responding with either "yes" or "no".
Answer:
"""

DEFAULT_VERIFY_PROMPT = """
Check if the given answer correctly reflects whether the claim is consistent with the corresponding document. Consistency in this context implies that
all information presented in the claim is substantiated by the document. If not, it should be considered inconsistent. "yes" means the claim is consistent with the document, and "no" means the claim is not consistent with the document.

Document: {source.doc}
Claim: {source.claim}
Answer: {previous.choice}
Answer Justification: {previous.explanation}

If you think the answer is correct, return the answer as is. If it's incorrect, return the opposite (if answer = "yes", return "no", and if answer = "no", return "yes").
"""


@dataclass
class EnsembleVerifyJudge:
    judge_prompt: str = DEFAULT_JUDGE_PROMPT
    verify_prompt: str = DEFAULT_VERIFY_PROMPT
    output_categories: DiscreteScale = field(
        default_factory=lambda: DiscreteScale(["yes", "no"])
    )
    model: str = "gpt-4o"
    repeat: int = 3
    retries: int = 3
    judge_temperature: float = 0.7
    verify_temperature: float = 0.0
    judge_explanation: bool = True
    pipeline: Pipeline = field(init=False)
    kwargs: dict = field(default_factory=dict)

    def __post_init__(self):
        self.hierarchical_protocol = (
            Block()
            >> CategoricalJudgeUnit(
                name="judge",
                categories=self.output_categories,
                explanation=self.judge_explanation,
            )
            .prompt(self.judge_prompt)
            .via(
                policy_or_name=self.model,
                retries=self.retries,
                temperature=self.judge_temperature,
                **self.kwargs,
            )
            >> CategoricalJudgeUnit(name="verify", categories=self.output_categories)
            .prompt(self.verify_prompt)
            .via(
                policy_or_name=self.model,
                retries=self.retries,
                temperature=self.verify_temperature,
                **self.kwargs,
            )
        )

        self.pipeline = (
            Pipeline()
            >> Layer([self.hierarchical_protocol], repeat=self.repeat)
            >> MaxPoolUnit()
        )

    def run(self, dataset: List[Schema]) -> List[str]:
        results_df, leaf_node_cols = self.pipeline.run_from_list(dataset)
        return list(results_df[leaf_node_cols[0]])


if __name__ == "__main__":
    judge = EnsembleVerifyJudge()
    test = judge.run(
        [
            # Examples from ExpertQA
            Schema.of(
                doc="DX Bash learning on the different ways on how they can make the child comfortable and relaxed. Find non-verbal cues to connect Communicating with a child with autism may be daunting. However, you need to talk or even touch to bond and communicate. The tone of your voice and body language have a significant impact on how you communicate with a child suffering from autism; hence you need to learn their language to help improve communications between you, this will make them feel loved and accepted within the family. Maintain a personalized autism treatment plan With many different treatments available, it can",
                claim="The child can feel loved and accepted within the family this way.",
            ),
            Schema.of(
                doc="Greece History and Culture | WorkingAbroad Greece - History and Culture Greek history stretches right back to the Paleolithic era, circa 400,000 - 13,000 BP. Other periods into which the country's history is usually sorted are Mesolithic (circa 10,000 - 7000 BCE, Neolithic (circa 7000 - 3000 BCE), Bronze Age (circa 3300 - 1150 BCE), Dark Ages (circa 1100 - 700 BCE), Archaic (circa 700 - 480 BCE), Classical (480 - 323 BCE) and Hellenistic (323 - 30 BCE). And these are all prior to the infamous Roman conquest of the region circa 146 BC-324 AD. After this, the area was still influenced by the rule",
                claim="However, it is important to note that the idea of hero cults and hero worship was a common practice earlier in Greek history, dating back to the Archaic period (circa 800-480 BCE).",
            ),
        ]
    )
    print(test)



import inspect
from abc import ABC
from collections import defaultdict
from typing import Any, Dict, List, Optional, Tuple, Type, Union, get_args

from loguru._logger import Logger
from pydantic import BaseModel, ConfigDict, Field, PrivateAttr, create_model
from pydantic.fields import FieldInfo

from verdict.scale import Scale
from verdict.util.exceptions import ConfigurationError
from verdict.util.misc import shorten_string


class Schema(BaseModel, ABC):
    model_config = ConfigDict(frozen=True, arbitrary_types_allowed=True)
    _scales: Dict[str, Scale] = PrivateAttr(default_factory=dict)

    @classmethod
    def get_scale(cls, field_name: str) -> Scale:
        return cls._scales[field_name]

    def __init_subclass__(cls) -> None:
        cls._scales = {}

        for field_name, field_type in cls.__annotations__.items():
            if inspect.isclass(field_type) and issubclass(field_type, Scale):
                if field_name not in cls.__dict__:
                    raise ConfigurationError(f"""No token support specified for Scale field '{field_name}'.

If using a Scale field type, you must specify its token support. For example,

```
class ResponseSchema(Schema):
    score: Scale = DiscreteScale((1, 5))
```
""")

                scale_instance: Scale = cls.__dict__[field_name]
                # Handle (scale, Field(...)) from an Schema.from_values() call
                if isinstance(scale_instance, tuple):
                    scale_instance = scale_instance[0]
                cls.__annotations__[field_name] = scale_instance.T

                cls._scales[field_name] = scale_instance
                setattr(
                    cls,
                    field_name,
                    scale_instance.pydantic_fields(field_name)[field_name][1],
                )

        BaseModel.__init_subclass__()

    @staticmethod
    def infer_pydantic_annotation(obj: Any) -> Any:
        """Infer the appropriate Pydantic annotation for a given object."""

        # Handle None
        if obj is None:
            return None

        # Handle direct types (int, str, etc.)
        if isinstance(obj, type):
            return obj

        obj_type = type(obj)
        if isinstance(obj, Schema):
            return Schema

        # Handle List-like objects
        if isinstance(obj, list) and obj:
            elem_type = Schema.infer_pydantic_annotation(obj[0])
            return List[elem_type]

        # Handle Tuple-like objects
        if isinstance(obj, tuple) and obj:
            elem_types = tuple(Schema.infer_pydantic_annotation(e) for e in obj)
            return Tuple[elem_types]

        # Handle Dict-like objects
        if isinstance(obj, dict) and obj:
            key_type = Schema.infer_pydantic_annotation(next(iter(obj.keys())))
            value_type = Schema.infer_pydantic_annotation(next(iter(obj.values())))
            return Dict[key_type, value_type]

        # Default: Return the object's direct type
        return obj_type

    @staticmethod
    def infer(**kwargs) -> Type["Schema"]:
        # Schema.infer(score=5) -> BaseModel(score=int)
        return type(
            "InferredSchema",
            (Schema,),
            {
                **{
                    field_name: field_value
                    for field_name, field_value in kwargs.items()
                    if isinstance(field_value, Scale)
                },
                "__module__": __name__,
                "__annotations__": {
                    field_name: Schema.infer_pydantic_annotation(field_value)
                    for field_name, field_value in kwargs.items()
                },
            },
        )

    @staticmethod
    def of(**kwargs) -> "Schema":
        # Schema.of(score=5) -> BaseModel(score=int)(score=5)
        return type(
            "InferredSchema",
            (Schema,),
            {
                **kwargs,
                "__module__": __name__,
                "__annotations__": {
                    field_name: Schema.infer_pydantic_annotation(field_value)
                    for field_name, field_value in kwargs.items()
                },
            },
        )(**kwargs)

    @staticmethod
    def inline(**kwargs) -> Type["Schema"]:
        # Schema.inline(score=int) -> BaseModel(score=int)
        return type(
            "InlineSchema",
            (Schema,),
            {
                "__module__": __name__,
                "__annotations__": {
                    field_name: field_type for field_name, field_type in kwargs.items()
                },
            },
        )

    @classmethod
    def empty(cls) -> "Schema":
        return type(
            "EmptySchema",
            (Schema,),
            {
                "__module__": __name__,
            },
        )()

    @classmethod
    def is_empty(cls) -> bool:
        return len(cls.model_fields) == 0

    def __str__(cls_or_self) -> str:
        if isinstance(cls_or_self, type):
            return str(cls_or_self.model_fields)
        return super().__str__()

    def __repr__(self) -> str:
        return self.__str__()

    @classmethod
    def append(cls, **kwargs) -> Type["Schema"]:
        model_fields = {
            **{
                field_name: (field_info.annotation, Field(...))
                for field_name, field_info in cls.model_fields.items()
            },
            **{
                field_name: (t if isinstance(t, tuple) else (t, Field(...)))
                for field_name, t in kwargs.items()
            },
        }

        return create_model(  # type: ignore
            cls.__name__, **model_fields, __base__=Schema
        )

    def add(self, **kwargs) -> "Schema":
        return type(self).append(
            **{
                field_name: Schema.infer_pydantic_annotation(field_value)
                for field_name, field_value in kwargs.items()
            }
        )(**{**self.model_dump(), **kwargs})

    @classmethod
    def prepend(cls, **kwargs) -> Type["Schema"]:
        model_fields = {
            **{
                field_name: (t if isinstance(t, tuple) else (t, Field(...)))
                for field_name, t in kwargs.items()
            },
            **{
                field_name: (field_info.annotation, Field(...))
                for field_name, field_info in cls.model_fields.items()
            },
        }

        return create_model(  # type: ignore
            cls.__name__, **model_fields, __base__=Schema
        )

    @staticmethod
    def from_values(**kwargs) -> "Schema":
        return create_model(  # type: ignore
            "Schema",
            **{
                name: (Schema.infer_pydantic_annotation(value), Field(...))
                for name, value in kwargs.items()
            },
            __base__=Schema,
        )(**kwargs)

    def escape(self) -> str:
        representation = str(self).replace("{", "{{").replace("}", "}}")
        return shorten_string(representation)

    def conform(
        self, expected: Type["Schema"], logger: Optional[Logger] = None
    ) -> "Schema":
        """
        Conform the current schema to the expected schema.
        For a missing field in this Schema (eg, in expected, but not in type(self)),
            1. Check if there is a default factory for the field in expected. If so, use that.
            2. Copy the first field in `self` that matches the type of the expected field.
        """
        current = self

        current_lookup = type(current)._fieldinfo_lookup()
        expected_lookup = expected._fieldinfo_lookup()

        # 1. Set defaults
        for expected_field_name, expected_field_info in expected.model_fields.items():
            key = Schema.generate_key(expected_field_info)

            # names + types match, so we're good
            if (
                expected_field_name in current.model_fields
                and key
                == Schema.generate_key(current.model_fields[expected_field_name])
            ):
                continue

            # default/default factory
            if not expected_field_info.is_required():
                current = current.add(
                    **{
                        expected_field_name: (
                            field_value := expected_field_info.default_factory()
                            if expected_field_info.default_factory is not None
                            else expected_field_info.default
                        )
                    }
                )

                current_lookup = type(current)._fieldinfo_lookup()
                if logger:
                    logger.info(
                        f"Constructed default input field {expected_field_name}={field_value} from {current}"
                    )

        # 2. Copy fields
        for expected_field_name, expected_field_info in expected.model_fields.items():
            key = Schema.generate_key(expected_field_info)

            # names + types match, so we're good
            if (
                expected_field_name in current.model_fields
                and key
                == Schema.generate_key(current.model_fields[expected_field_name])
            ):
                continue

            if len(current_lookup[key]) < len(expected_lookup[key]):
                raise ConfigurationError(
                    f"Cannot cast input. Not enough fields with FieldInfo: {expected_field_info.__str__()} (found {len(current_lookup[key])}, but we need to populate {len(expected_lookup[key])})"
                )

            # find a compatible field in the current schema
            found: bool = False
            for field_name, field_info in current.model_fields.items():
                if key == Schema.generate_key(field_info):
                    current = current.add(
                        **{expected_field_name: getattr(current, field_name)}
                    )
                    current_lookup = type(current)._fieldinfo_lookup()
                    if logger:
                        logger.info(
                            f"Copied field {field_name}={getattr(current, field_name)} to {expected_field_name}"
                        )
                    found = True
                    break

            if found:
                continue

            raise ConfigurationError(
                f"Cannot conform {self} to {expected}. Field compatible with type '{field_info.__str__()}' not found in the input model."
            )

        return current

    @staticmethod
    def generate_key(field_info: FieldInfo) -> str:
        def remove_optional(typ: Type) -> Type:
            if hasattr(typ, "__origin__") and typ.__origin__ is Union:
                args = tuple(arg for arg in get_args(typ) if arg is not type(None))
                if len(args) == 1:
                    return args[0]
                return Union[args]
            return typ

        ATTRIBUTES_FOR_EQUIVALENCE = [
            "le",
            "ge",
            "gt",
            "lt",
            "eq",
            "ne",
            "ge",
            "le",
            "metadata",
        ]  # metadata includes pattern
        keys = [f"{remove_optional(field_info.annotation)}"]
        keys += [
            f"{k}={v}"
            for k, v in zip(
                ATTRIBUTES_FOR_EQUIVALENCE,
                map(
                    lambda attr: getattr(field_info, attr, None),
                    ATTRIBUTES_FOR_EQUIVALENCE,
                ),
            )
        ]
        return str(hash(",".join(keys)))

    @classmethod
    def _fieldinfo_lookup(cls) -> Dict[str, Type]:
        lookup = defaultdict(list)
        for field_name, field_info in cls.model_fields.items():
            lookup[Schema.generate_key(field_info)].append((field_name, field_info))
        return lookup



from dataclasses import dataclass, field
from typing import List

from verdict import Pipeline
from verdict.common.cot import CoTUnit
from verdict.common.judge import JudgeUnit
from verdict.scale import DiscreteScale
from verdict.schema import Schema

DEFAULT_JUDGE_PROMPT = """
Based on the following, extract the final score.

{previous.thinking}

Score:
"""


DEFAULT_G_EVAL_COHERENCE_PROMPT = """
You will be given one summary written for a news article.

Your task is to rate the summary on one metric.

Please make sure you read and understand these instructions carefully. Please keep this
document open while reviewing, and refer to it as needed.

Evaluation Criteria:

Coherence (1-5) - the collective quality of all sentences. We align this dimension with
the DUC quality question of structure and coherence whereby ”the summary should be
well-structured and well-organized. The summary should not just be a heap of related information, but should build from sentence to sentence to a coherent body of information about a topic.”

Evaluation Steps:

1. Read the news article carefully and identify the main topic and key points.
2. Read the summary and compare it to the news article. Check if the summary covers the main
topic and key points of the news article, and if it presents them in a clear and logical order.
3. Assign a score for coherence on a scale of 1 to 5, where 1 is the lowest and 5 is the highest
based on the Evaluation Criteria.

News Article:
{source.document}

Summary:
{source.summary}

Evaluation Form (scores ONLY):
- Coherence:
"""


@dataclass
class GEvalJudge:
    cot_prompt: str = DEFAULT_G_EVAL_COHERENCE_PROMPT
    cot_model: str = "gpt-4o"
    cot_temperature: float = 0.5
    judge_prompt: str = DEFAULT_JUDGE_PROMPT
    judge_model: str = "gpt-4o-mini"
    judge_temperature: float = 0.0
    retries: int = 3
    kwargs: dict = field(default_factory=dict)
    pipeline: Pipeline = field(init=False)

    def __post_init__(self):
        self.pipeline = (
            Pipeline()
            >> CoTUnit(name="GEval-CoT")
            .prompt(self.cot_prompt)
            .via(
                policy_or_name=self.cot_model,
                retries=self.retries,
                temperature=self.cot_temperature,
                **self.kwargs,
            )
            >> JudgeUnit(name="GEval-Judge", scale=DiscreteScale((1, 5)))
            .prompt(self.judge_prompt)
            .via(
                policy_or_name=self.judge_model,
                retries=self.retries,
                temperature=self.judge_temperature,
                **self.kwargs,
            )
        )

    def run(self, dataset: List[Schema]) -> List[float]:
        results_df, leaf_node_cols = self.pipeline.run_from_list(dataset)
        return list(results_df[leaf_node_cols[0]])


if __name__ == "__main__":
    judge = GEvalJudge()
    test = judge.run(
        [
            Schema.of(
                document="HackerOne: Security researchers I've worked with are blown away by the speed, quality, and variety of prompts Haize Labs generates for their attacks. Its intuitive user interface simplifies the process of designing and executing tests, while the robust reporting and analytics features help turn raw findings into actionable insights. For researchers dedicated to advancing AI reliability, Haize Labs delivers a powerful solution that makes testing smarter and more rewarding.",
                summary="Haize Labs has built out a powerful platform to red-team AI quickly, cheaply, and efficiently. Its platform features an intuitive UI that makes testing, reporting of, and improving AI systems incredibly easy. The Haize product is a must-use for researchers looking to up their eval game.",
            ),
            Schema.of(
                document="HackerOne: Security researchers I've worked with are blown away by the speed, quality, and variety of prompts Haize Labs generates for their attacks. Its intuitive user interface simplifies the process of designing and executing tests, while the robust reporting and analytics features help turn raw findings into actionable insights. For researchers dedicated to advancing AI reliability, Haize Labs delivers a powerful solution that makes testing smarter and more rewarding.",
                summary=" buidling fdas dfsfdasfdkasokfdpsa by haize Labs that is powerful  to team wit hred Intentions, also with quickness, cheapness, efficientness, Intelligence that is Artificial.",
            ),
        ]
    )
    print(test)



from verdict.prompt import Prompt


class GEvalEvaluationStepsCoTPrompt(Prompt):
    """
    {root.task}

    Evaluation Criteria:
    {criteria.name} ({criteria.scale}) - {criteria.description}

    Evaluation Steps:
    """



import string
from typing import List, Optional

from verdict.core.primitive import Unit
from verdict.scale import DiscreteScale
from verdict.schema import Field, Schema


class RankerUnit(Unit):
    _char: str = "Ranker"

    class InputSchema(Schema):
        # need a sentinel for empty, but using [''] for now makes type inference simple
        options: List[str] = Field(default=[""])

    class ResponseSchema(Schema):
        ranking: List[str]

    class OutputSchema(Schema):
        ranked_options: List[str]

    def __init__(
        self,
        k: int = 3,
        options: Optional[DiscreteScale] = None,
        explanation: bool = False,
        original: bool = False,
        **kwargs,
    ):
        self.k = k
        self.explanation = explanation
        self.original = original

        options = options or DiscreteScale(list(string.ascii_uppercase[:k]))
        self.scale = options

        if explanation:
            self.ResponseSchema = self.ResponseSchema.prepend(explanation=str)  # type: ignore
            self.OutputSchema = self.OutputSchema.prepend(explanation=str)  # type: ignore

        super().__init__(**kwargs)

    def validate(self, input: InputSchema, response: ResponseSchema) -> bool:
        if self.original:
            assert len(input.options) != [""], (
                "Pass the options in InputSchema.options to index original choices."
            )
            assert len(input.options) == self.k, (
                f"Number of input options must equal {self.k}"
            )

    def process(self, input: InputSchema, response: ResponseSchema) -> OutputSchema:
        if not self.original:
            return response

        fields = {
            "ranked_options": [
                input.options[
                    self.scale.values.index(self.scale.value_mapping_fn(rank))
                ]
                for rank in response.ranking
            ],
        }

        if self.explanation:
            fields["explanation"] = response.explanation

        return self.OutputSchema(**fields)



import os
from dataclasses import dataclass
from pathlib import Path
from typing import Any, Dict

from verdict.model import ModelSelectionPolicy
from verdict.util.ratelimit import (
    ConcurrentRateLimiter,
    RateLimitPolicy,
    TimeWindowRateLimiter,
)


# Global state
@dataclass
class Config:
    rate_limiter_disabled: bool = False


state = Config()

# Defaults
## Rate limiting
PROVIDER_RATE_LIMITER: Dict[str, RateLimitPolicy] = {
    "deepinfra": RateLimitPolicy(
        {ConcurrentRateLimiter(max_concurrent=200): "requests"}
    ),
    "together_ai": RateLimitPolicy.of(rpm=600, tpm=180_000),
    "openai": RateLimitPolicy(
        {  # tier 1 for gpt-4o-mini
            TimeWindowRateLimiter(max_value=5, window_seconds=60): "requests",
            TimeWindowRateLimiter(
                max_value=10_000, window_seconds=60 * 60 * 24
            ): "requests",
            TimeWindowRateLimiter(max_value=200_000, window_seconds=60): "tokens",
        }
    ),
    # TODO: add other providers
}

DEFAULT_RATE_LIMITER: RateLimitPolicy = PROVIDER_RATE_LIMITER["openai"]

## Connection parameters
DEFAULT_PROVIDER_TIMEOUT: int = 120
DEFAULT_PROVIDER_STREAM_TIMEOUT: int = 120

## Inference parameters
DEFAULT_MODEL_SELECTION_POLICY: ModelSelectionPolicy = ModelSelectionPolicy.from_name(
    "gpt-4o-mini", retries=3
)
DEFAULT_INFERENCE_PARAMS: dict[str, Any] = {}

## Token extraction prompt
TOKEN_EXTRACTOR_SPECIFICATION_PROMPT: str = """
{content}
--------------------------------

Do not explain your answer. {scale_prompt}
"""

# System
## Logging
DEBUG: bool = bool(os.getenv("DEBUG", False))

LIGHTWEIGHT_EXECUTOR_WORKER_COUNT: int = 32

VERDICT_LOG_DIR: Path = Path.cwd() / ".verdict"
VERDICT_LOG_DIR.mkdir(parents=True, exist_ok=True)



import pprint
import textwrap
from abc import ABC, abstractmethod
from contextlib import contextmanager
from dataclasses import dataclass, field
from typing import (
    Any,
    Callable,
    ContextManager,
    Dict,
    Iterator,
    List,
    Optional,
    Tuple,
    Union,
)

from loguru._logger import Logger
from typing_extensions import Self

from verdict.schema import Schema
from verdict.util.exceptions import ConfigurationError
from verdict.util.misc import DisableLogger, shorten_string
from verdict.util.ratelimit import (
    RateLimitConfig,
    RateLimitPolicy,
    UnlimitedRateLimiter,
)


class Model(ABC):
    name: str
    use_nonce: bool = False  # Default value for the base class

    rate_limiter: Optional[Union[RateLimitPolicy, RateLimitConfig]] = field(
        repr=False, default=None
    )
    rate_limit: RateLimitPolicy = field(init=False, repr=True)

    def __post_init__(self) -> None:
        object.__setattr__(self, "provider", self.name.split("/")[0])

    def _configure_rate_limiter(self) -> None:
        from verdict import config

        if config.state.rate_limiter_disabled:
            rate_limit = RateLimitPolicy(
                {
                    UnlimitedRateLimiter(): "requests",
                }
            )
        elif self.rate_limiter is None:
            rate_limit = config.DEFAULT_RATE_LIMITER
        elif isinstance(self.rate_limiter, dict):
            rate_limit = RateLimitPolicy(self.rate_limiter)
        elif isinstance(self.rate_limiter, RateLimitPolicy):
            rate_limit = self.rate_limiter
        else:
            raise ConfigurationError(
                "Invalid rate_limiter passed. Must be a RateLimit or a dictionary of RateLimiters -> metric"
            )

        object.__setattr__(self, "rate_limit", rate_limit)

    @property
    def char(self) -> str:
        return self.name.split("/")[-1]

    def __str__(self) -> str:
        components = self.name.split("/")
        if len(components) == 1:
            return f"{self.__class__.__name__}({self.name})"
        return f"{self.__class__.__name__}(.../{components[-1]})"

    @property
    def connection_parameters(self) -> dict[str, Any]:
        return {
            "num_retries": 1,  # NOTE: this overrides litellm's retry mechanism so we can control it with ModelSelectionPolicy
            "max_retries": 1,  # NOTE: ditto, but for instructor
        }


@dataclass(frozen=True)
class ProviderModel(Model):
    name: str
    use_nonce: bool = False

    rate_limiter: Optional[Union[RateLimitPolicy, RateLimitConfig]] = field(
        repr=False, default=None
    )
    rate_limit: RateLimitPolicy = field(init=False, repr=True)

    def __post_init__(self) -> None:
        super().__post_init__()

        object.__setattr__(self, "use_nonce", True)

        from litellm import get_llm_provider  # type: ignore[import-untyped]

        _, provider, _, _ = get_llm_provider(self.name)
        object.__setattr__(self, "provider", provider)

        from verdict import config

        if self.rate_limiter is None:
            rate_limiter = config.PROVIDER_RATE_LIMITER.get(
                provider, config.DEFAULT_RATE_LIMITER
            )
            object.__setattr__(self, "rate_limiter", rate_limiter)

        self._configure_rate_limiter()

    @property
    def connection_parameters(self) -> dict[str, Any]:
        from verdict import config

        return {
            "model": self.name,
            "timeout": config.DEFAULT_PROVIDER_TIMEOUT,  # NOTE: this can be overridden by the inference_parameters passed to the Client
            "stream_timeout": config.DEFAULT_PROVIDER_STREAM_TIMEOUT,  # NOTE: this is set to `timeout` if streaming is True
            **super().connection_parameters,
        }


@dataclass(frozen=True)
class vLLMModel(Model):
    name: str
    api_base: str
    api_key: str
    use_nonce: bool = field(init=False)

    rate_limiter: Optional[Union[RateLimitPolicy, RateLimitConfig]] = field(
        repr=False, default=None
    )
    rate_limit: RateLimitPolicy = field(init=False, repr=True)

    def __post_init__(self) -> None:
        super().__post_init__()

        object.__setattr__(self, "use_nonce", False)
        self._configure_rate_limiter()

    @property
    def connection_parameters(self) -> dict[str, Any]:
        return {
            "model": f"hosted_vllm/{self.name}",
            "api_base": self.api_base,
            "api_key": self.api_key,
            "timeout": None,  # NOTE: this can be overridden by the inference_parameters passed to the Client
            "stream_timeout": None,  # NOTE: this is set to `timeout` if streaming is True
            **super().connection_parameters,
        }


@dataclass
class Client:
    complete: Callable
    model: Model
    inference_parameters: dict[str, Any]

    def _supports_image_input(self) -> bool:
        if hasattr(self.model, "provider"):
            # TODO: add more model providers
            VISION_MODELS = {
                "openai": [
                    "o4-mini",
                    "o3",
                    "o3-pro",
                    "o1",
                    "o1-pro",
                    "gpt-4.1",
                    "gpt-4o",
                    "chatgpt-4o-latest",
                    "gpt-4.1-mini",
                    "o4-mini",
                    "gpt-4.1-nano",
                    "gpt-4o-mini",
                ],
            }

            provider_models = VISION_MODELS.get(self.model.provider, [])
            return self.model.char in provider_models
        return False

    def _has_image_content(
        self, messages: List[Dict[str, str | List[Dict[str, str | Dict[str, str]]]]]
    ) -> bool:
        for message in messages:
            content = message.get("content", "")
            if isinstance(content, list):
                for item in content:
                    if isinstance(item, dict) and item.get("type") == "image_url":
                        return True

        return False

    @contextmanager
    def defaults(self, **inference_parameters_defaults) -> ContextManager[None]:  # type: ignore
        original_inference_parameters = self.inference_parameters

        self.inference_parameters = {
            **inference_parameters_defaults,
            **original_inference_parameters,
        }
        yield

        self.inference_parameters = original_inference_parameters

    def __call__(
        self,
        logger: Logger,
        messages: List[Dict[str, str]],
        response_model: Optional[Schema] = None,
        streaming: bool = False,
    ):
        from verdict import config

        if self._has_image_content(messages) and not self._supports_image_input():
            raise ConfigurationError(
                f"Model {self.model.name} does not support image inputs. "
                f"Images were detected in the message content. "
            )

        logger.debug(
            textwrap.dedent(
                f"""
                Preparing parameters for {repr(self.model)} with
                specified connection_parameters: {self.model.connection_parameters}
                default inference_parameters: {config.DEFAULT_INFERENCE_PARAMS}
                specified inference_parameters: {self.inference_parameters}
                """
            )
        )

        # 1. add in messages, connection_parameters, and inference_parameters (defaults first)
        parameters = {
            "messages": messages,
            **self.model.connection_parameters,
            **config.DEFAULT_INFERENCE_PARAMS,
            **self.inference_parameters,
        }

        # 2. configure streaming
        if streaming:
            parameters["stream"] = streaming
            if response_model is not None:
                # needed for streaming structured output
                from instructor import Partial  # type: ignore[import-untyped]

                response_model = Partial[response_model]
            if "timeout" in self.inference_parameters:
                parameters["stream_timeout"] = self.inference_parameters["timeout"]

        # 3. configure structured output
        if response_model is not None:
            parameters["response_model"] = response_model

        parameters_no_api_key = {
            p: v for p, v in parameters.items() if p not in ["api_key"]
        }
        logger.debug(
            f"Sending parameters for {repr(self.model)}:  {pprint.pformat(parameters_no_api_key, width=80, depth=3)}"
        )
        return self.complete(**parameters)


class ClientWrapper:
    model: Model
    inference_parameters: dict[str, Any]

    def __init__(self, model: Model, **inference_parameters) -> None:
        self.model = model
        self.model._configure_rate_limiter()

        self.inference_parameters = inference_parameters

        import litellm

        litellm.drop_params = True
        litellm.suppress_debug_info = True
        self.raw_client = Client(
            litellm.LiteLLM().chat.completions.create,
            self.model,
            self.inference_parameters,
        )

        from instructor import Mode, patch  # type: ignore[import-untyped]

        MODEL_PROVIDER_TO_INSTRUCTOR_MODE_OVERRIDE = {
            "huggingface": Mode.JSON,
            "deepinfra": Mode.JSON,
            "together_ai": Mode.JSON,
        }

        MODEL_TO_INSTRUCTOR_MODE_OVERRIDE = {
            "openai/o1": Mode.JSON_O1,
            "openai/o1-2024-12-17": Mode.JSON_O1,
        }

        with DisableLogger("LiteLLM"):
            self.function_calling_client = Client(
                patch(
                    litellm.LiteLLM(),
                    mode=MODEL_TO_INSTRUCTOR_MODE_OVERRIDE.get(
                        f"{self.model.provider}/{self.model.char}",
                        MODEL_PROVIDER_TO_INSTRUCTOR_MODE_OVERRIDE.get(
                            self.model.provider, Mode.TOOLS
                        ),
                    ),
                ).chat.completions.create,  # type: ignore
                self.model,
                self.inference_parameters,
            )

    def encode(self, word: str) -> List[int]:
        import tokenizers  # type: ignore[import-untyped]
        from litellm import encode  # type: ignore[import-untyped]

        tokens = encode(model=self.model.name, text=shorten_string(word, encoded=False))
        if isinstance(tokens, tokenizers.Encoding):
            return tokens.ids

        return tokens

    @staticmethod
    def from_model(model: Model, **inference_parameters) -> "ClientWrapper":
        return ClientWrapper(model, **inference_parameters)


class ModelSelectionPolicy:
    @property
    def char(self) -> str:
        return str(self)

    def __init__(self) -> None:
        self.client_factories: List[Callable[[], ClientWrapper]] = []
        self.client_configs: List[Tuple[Model, int, dict]] = []
        self.client_repeats: List[Tuple[Model, int]] = []

    def __len__(self) -> int:
        return len(self.client_factories)

    def get_clients(self) -> Iterator[ClientWrapper]:
        return iter(map(lambda factory: factory(), self.client_factories))

    @staticmethod
    def from_name(
        model: Union[str, Model], retries: int = 1, **inference_parameters
    ) -> "ModelSelectionPolicy":
        policy = ModelSelectionPolicy()
        for _ in range(retries):
            if isinstance(model, str):
                model = ProviderModel(name=model)

            policy.client_factories.append(
                lambda: ClientWrapper.from_model(model, **inference_parameters)
            )
            policy.client_configs.append((model, retries, inference_parameters))
        policy.client_repeats.append((model, retries))

        return policy

    @staticmethod
    def from_names(model_names: List[Tuple[str, int, dict]]) -> "ModelSelectionPolicy":
        policy = ModelSelectionPolicy()
        for model_name, retries, inference_parameters in model_names:
            policy += ModelSelectionPolicy.from_name(
                model_name, retries, **inference_parameters
            )

        return policy

    @staticmethod
    def from_any(
        policy_or_name: Union[
            "ModelSelectionPolicy", str, Model, List[Union[str, Model]]
        ],
        retries: int = 1,
        **inference_parameters,
    ) -> "ModelSelectionPolicy":
        if isinstance(policy_or_name, ModelSelectionPolicy):
            return policy_or_name
        elif isinstance(policy_or_name, (str, Model)):
            return ModelSelectionPolicy.from_name(
                policy_or_name, retries, **inference_parameters
            )
        elif isinstance(policy_or_name, list) and len(policy_or_name) > 0:
            if isinstance(policy_or_name[0], (str, Model)):
                return ModelSelectionPolicy.from_names(
                    [(name, retries, inference_parameters) for name in policy_or_name]
                )
            elif isinstance(policy_or_name[0], tuple):
                return ModelSelectionPolicy.from_names(policy_or_name)
            else:
                raise ConfigurationError(
                    "Invalid argument type. Expected str, list, or ModelSelectionPolicy."
                )
        else:
            raise ConfigurationError(
                "Invalid argument type. Expected str, list, or ModelSelectionPolicy."
            )

    def __add__(self, other: "ModelSelectionPolicy") -> Self:
        self.client_factories.extend(other.client_factories)
        self.client_configs.extend(other.client_configs)
        self.client_repeats.extend(other.client_repeats)
        return self

    def __str__(self) -> str:
        return ", ".join(
            f"{retries}x {model.name}" for model, retries in self.client_repeats
        )

    def __repr__(self) -> str:
        return self.__str__()


class ModelConfigurable(ABC):
    _model_selection_policy: Optional[ModelSelectionPolicy] = None

    def via(
        self,
        policy_or_name: Union[
            ModelSelectionPolicy, str, Model, List[Union[str, Model]]
        ],
        retries: int = 1,
        **inference_parameters,
    ) -> Self:
        self.set(
            "model_selection_policy",
            ModelSelectionPolicy.from_any(
                policy_or_name, retries, **inference_parameters
            ),
        )
        return self

    # provided by Node, Graph
    @abstractmethod
    def set(self, attr, value) -> None:
        pass

    @property
    def model_selection_policy(self) -> Optional[ModelSelectionPolicy]:
        return self._model_selection_policy

    @model_selection_policy.setter
    def model_selection_policy(
        self, model_selection_policy: ModelSelectionPolicy
    ) -> None:
        from verdict.util.log import logger

        # only set the model selection policy if it hasn't been set yet
        if self._model_selection_policy is None:
            self._model_selection_policy = model_selection_policy
            logger.info(f"Model selection policy set to {model_selection_policy.char}")
        else:
            logger.warning("Model selection policy skipped. Already set.")
            pass



# high-level
class VerdictSystemError(Exception):
    """
    Non user-defined errors that are related to the core Verdict framework.
    """


class VerdictDeclarationTimeError(Exception):
    """
    Raised for a deterministic (i.e., cannot be fixed with a retry) error that
    occurs as a result of a configuration error.

    Examples:
    * using an incompatible Extractor with a particular Scale
    * invalid keys in a Prompt
    """


class VerdictExecutionTimeError(Exception):
    """
    Raised for a non-deterministic (i.e., can be fixed with a retry) error that
    occurs at execution time.

    Examples:
    * provider error
        * rate-limit
        * function calling validation
        * timeout
    """


# Declaration-time errors
class InputSchemaMismatchError(VerdictDeclarationTimeError):
    pass


class PromptError(VerdictDeclarationTimeError):
    pass


class ConfigurationError(VerdictDeclarationTimeError):
    """
    User error with the configuration of a Verdict pipeline.
    """


class PostProcessError(VerdictDeclarationTimeError):
    """
    Indicates an error in the process() method of a Unit.
    """

    pass


class PropagateError(VerdictDeclarationTimeError):
    """
    Indicates an error in the passed propagator function.
    """

    pass


# Execution-time errors
class StructuredOutputError(VerdictExecutionTimeError):
    pass


class ProviderError(VerdictExecutionTimeError):
    pass


class PostValidationError(VerdictExecutionTimeError):
    pass



from __future__ import annotations

import logging
import re
import signal
import sys
import threading
from functools import wraps
from typing import Any, Callable, Optional, Type


def is_signal_safe():
    """
    Strictly check if the current environment is safe for signal handling.

    Returns:
        bool: True only if ALL signal safety conditions are met
    """
    # Must be in the main thread
    if threading.current_thread() is not threading.main_thread():
        return False

    # Must be in the main interpreter
    if not hasattr(sys, "argv"):
        return False

    # Additional interpreter state checks
    if not sys.modules.get("__main__"):
        return False

    # Check if the interpreter is shutting down
    if sys is None or threading is None:
        return False

    return True


def keyboard_interrupt_safe(func: Callable[[Any], Any]) -> Callable[[Any], Any]:
    @wraps(func)
    def wrapped(self: Any, *args, **kwargs) -> Any:
        if is_signal_safe():
            original_handler = signal.getsignal(signal.SIGINT)
            signal.signal(
                signal.SIGINT, lambda signum, frame: self.executor.graceful_shutdown()
            )

            try:
                return func(self, *args, **kwargs)
            except KeyboardInterrupt:
                self.executor.graceful_shutdown()
            finally:
                signal.signal(signal.SIGINT, original_handler)
        else:
            return func(self, *args, **kwargs)

    return wrapped


class DisableLogger:
    def __init__(self, logger_name: str, all: bool = False) -> None:
        self.logger = logging.getLogger(logger_name)
        self.previous_level: Optional[int] = None
        self.all = all

    def __enter__(self) -> None:
        self.previous_level = self.logger.level
        new_level = logging.CRITICAL
        if self.all:
            new_level += 1
        self.logger.setLevel(new_level)

    def __exit__(self, exc_type, exc_value, traceback) -> None:
        self.logger.setLevel(self.previous_level)  # type: ignore


def lightweight(cls: Type) -> Type:
    cls.lightweight = True
    return cls


def shorten_string(text: str, truncate_length: int = 10, encoded: bool = True) -> str:
    """
    Shorten inputs for logging, particularly if it contains base64 data.

    Args:
        text: The string that may contain base64 data
        truncate_length: How many characters to keep before truncating

    Returns:
        The string with base64 data truncated
    """

    def truncate_base64(match):
        base64_data = match.group(1)
        return f"'data': '{base64_data[: min(truncate_length, len(base64_data))]}...'"

    if encoded:
        # Check if the string contains image type patterns
        if re.search(r"'type': 'image/[^']*'", text):
            return re.sub(r"'data': '([^\s']{10,})'", truncate_base64, text)

        return text
    else:
        return re.sub(r"([^\s']{100,})", truncate_base64, text)



import copy
import operator
import time
from abc import ABCMeta
from dataclasses import dataclass, field
from enum import Enum
from itertools import cycle
from typing import (
    Callable,
    Dict,
    Generic,
    Iterator,
    List,
    Optional,
    Set,
    Tuple,
    Type,
    TypeVar,
    Union,
    final,
)

import numpy as np
import rich.align
import rich.console
import rich.layout
import rich.live
import rich.padding
import rich.panel
import rich.progress
import rich.spinner
import rich.table
import rich.text
import rich.tree
from loguru._logger import Logger
from typing_extensions import Self

import verdict.config as config
from verdict.core.executor import ExecutionState, Graph, Node, Task
from verdict.core.synchronization import SynchronizationState, UserState
from verdict.core.visualization import BranchManager, StreamingLayoutManager
from verdict.extractor import Extractor
from verdict.model import ModelConfigurable, ModelSelectionPolicy
from verdict.prompt import Promptable, PromptMessage
from verdict.schema import Schema
from verdict.util.exceptions import (
    ConfigurationError,
    PostProcessError,
    PostValidationError,
    PropagateError,
    VerdictDeclarationTimeError,
    VerdictExecutionTimeError,
)
from verdict.util.log import logger as base_logger
from verdict.util.misc import DisableLogger, shorten_string
from verdict.util.tracing import ExecutionContext


class Previous:
    def __init__(self, dependencies: Set["Unit"]) -> None:
        self.dependencies = sorted(
            list(dependencies), key=lambda node: getattr(node, "_ordering_timestamp", 0)
        )

    def __getattr__(self, key: str) -> Union[Schema, List[Schema]]:
        # e.g., previous.score -> (previous Judge unit dependency).output.score
        if len(self.dependencies) == 1 and hasattr(
            (unit := self.dependencies[0]).output, key
        ):
            return getattr(unit.output, key)

        # e.g., previous.judge -> (previous Judge unit dependency).output
        if (shortname := key.split(".")[0].lower()) in Unit._registry:
            _type = Unit._registry[shortname]
            matches = list(
                filter(
                    lambda node: isinstance(node, _type)
                    or node.__class__.__name__ == _type.__name__,
                    self.dependencies,
                )
            )

            # single dependency
            if len(matches) == 1:
                return matches[0].output

            # multiple dependencies
            return list(map(lambda unit: unit.output, matches))
        else:
            raise ConfigurationError(f"Previous {key.lower()} ({key}) not found")


@dataclass
class Propagator:
    fn: Callable[["Unit", Previous, Schema, Schema], Schema]

    @staticmethod
    def from_fn(
        fn: Callable[["Unit", Previous, Schema, Schema], Schema],
    ) -> "Propagator":
        return Propagator(fn=fn)

    def __call__(
        self, unit: "Unit", previous: Previous, source: Schema, output: Schema
    ) -> Schema:
        if not isinstance(previous, Previous):
            previous = Previous(previous)  # type: ignore

        return self.fn(unit, previous, source, output)

    def __copy__(self) -> "Propagator":
        return Propagator(fn=self.fn)

    @staticmethod
    def default() -> "Propagator":
        return Propagator.from_fn(lambda unit, previous, source, output: output)


class UnitRegistry(ABCMeta):
    _registry: Dict[str, Type["Unit"]] = {}

    def __new__(mcls, name, bases, namespace, **kwargs) -> Type["Unit"]:
        klass = super().__new__(mcls, name, bases, namespace, **kwargs)

        if name != "Unit":
            if "InputSchema" not in namespace and not any(
                hasattr(base, "InputSchema") for base in bases
            ):
                klass.InputSchema = Schema.inline()

            if "ResponseSchema" not in namespace and not any(
                hasattr(base, "ResponseSchema") for base in bases
            ):
                raise ConfigurationError(
                    f"Custom Unit '{name}' must define a 'ResponseSchema' class."
                )

            if "OutputSchema" not in namespace and not any(
                hasattr(base, "OutputSchema") for base in bases
            ):
                klass.OutputSchema = klass.ResponseSchema

        if name and not name.startswith("_") and name != "Unit":
            shortname = name.replace("Unit", "").lower()
            if shortname in mcls._registry:
                base_logger.warning(
                    f"Unit '{name}' (shortname='{shortname}') name collision in UnitRegistry. Replacing existing unit."
                )
            mcls._registry[shortname] = klass

        return klass


@dataclass
class MaterializationContext:
    root_block: Optional["Block"] = None
    prefix: List[str] = field(default_factory=lambda: ["root"])

    parent_branch: Optional[rich.tree.Tree] = None
    streaming_layout_manager: Optional[StreamingLayoutManager] = None

    n_instances: int = 1

    def copy(self) -> "MaterializationContext":
        return MaterializationContext(
            root_block=self.root_block,
            prefix=self.prefix.copy(),
            parent_branch=self.parent_branch,
            streaming_layout_manager=self.streaming_layout_manager,
            n_instances=self.n_instances,
        )

    def add_prefix(self, prefix: str) -> "MaterializationContext":
        self.prefix.append(prefix)
        return self


InputSchemaT = TypeVar("InputSchemaT", bound=Schema)
ResponseSchemaT = TypeVar("ResponseSchemaT", bound=Schema)
OutputSchemaT = TypeVar("OutputSchemaT", bound=Schema)


class DataFlowSchema(Generic[InputSchemaT, ResponseSchemaT, OutputSchemaT]):
    InputSchema: InputSchemaT
    ResponseSchema: ResponseSchemaT
    OutputSchema: OutputSchemaT


class Unit(
    Node,
    Task,
    ModelConfigurable,
    Promptable,
    DataFlowSchema[InputSchemaT, ResponseSchemaT, OutputSchemaT],
    metaclass=UnitRegistry,
):
    lightweight: bool = False

    shared: SynchronizationState  # fresh on init/copy, shared on clone
    data: UserState  # fresh on init, copied on a copy/clone
    model_selection_policy: Optional["ModelSelectionPolicy"] = None

    @property
    def _char(self) -> str:
        return "Unit"

    @final
    @property
    def char(self) -> str:
        model_selection_policy = (
            self.model_selection_policy or config.DEFAULT_MODEL_SELECTION_POLICY
        )
        _char = f"{self._char}"
        if not self.lightweight:
            _char += f" via {self.extractor.format().format(model_name=model_selection_policy.char)}"
        return _char

    def __init__(self, **kwargs):
        Node.__init__(self, **kwargs)
        Task.__init__(self)

        self._ordering_timestamp = time.perf_counter()
        self.shared = SynchronizationState()
        self.data = UserState()

        self._propagator = Propagator.default()

    def copy(self) -> Self:
        new = copy.copy(self)  # shallow (reference) copy

        new._ordering_timestamp = time.perf_counter()
        new.shared = SynchronizationState()
        new.data = (
            self.data.copy()
        )  # 2-level shallow copy (container reference + all values are copy.copy'd)

        return new

    def clone(self) -> Self:
        new = self.copy()
        new.shared = self.shared.add(new)

        return new

    @property
    def description(self) -> Optional[str]:
        return None

    def _materialize(
        self, context: MaterializationContext
    ) -> Tuple[Self, MaterializationContext]:
        context.add_prefix(f"unit[{self._char}{f' {self.name}' if self.name else ''}]")
        self.prefix = context.prefix

        with self.shared.lock:
            if self.shared.branch is None:
                self.shared.branch = BranchManager(
                    self.shared.branch_lock,
                    context.parent_branch,
                    context.n_instances,
                    self.should_pin_output,
                )
        self.shared.branch.update(ExecutionState.WAITING_FOR_DEPENDENCIES, self)

        self.streaming_layout_manager = context.streaming_layout_manager
        if context.root_block is not None:
            context.root_block.add(self.clear_dependencies())
        return self, context

    def __rshift__(self, other: Union["Unit", "Layer", "Block"]):  # type: ignore
        return (Block() >> self) >> other

    def link(self, other: Union["Unit", Graph["Unit"]]) -> None:  # type: ignore
        if self is other:
            return

        if isinstance(other, Unit):
            self.graph.setup_link(self, other)
            return

        with other.freeze():
            for node in other.root_nodes:
                self.link(node)

    def populate_prompt_message(self, input: Schema, logger: Logger) -> PromptMessage:
        return self._prompt.format(
            input_schema=input,
            unit=self,
            previous=Previous(self.dependencies),  # type: ignore
            source=self.source_input,
            logger=logger,
        )

    def execute(
        self, input: InputSchemaT, execution_context: Optional[ExecutionContext] = None
    ) -> OutputSchemaT:
        """
        Execute the unit with tracing.

        Args:
            input: The input schema.
            execution_context: The execution context for this execution (tracing, IDs, etc).

        Returns:
            The output schema.
        """
        if execution_context is None:
            execution_context = ExecutionContext()
        call_name = (
            getattr(self, "_char", None)
            or getattr(self, "char", None)
            or self.__class__.__name__
        )
        with execution_context.trace_call(
            name=call_name,
            inputs={"input": input, "unit": self},
        ) as call:
            logger = base_logger.bind(
                thread_id=self.thread_id, unit=".".join(self.prefix)
            )
            logger.info("Started Unit.execute()")

            if self.model_selection_policy is None:
                self.model_selection_policy = config.DEFAULT_MODEL_SELECTION_POLICY
                logger.debug(
                    f"Using default model selection policy: {self.model_selection_policy}"
                )

            if self.should_stream_output and self.streaming_layout_manager:
                streaming_layout = self.streaming_layout_manager.add(self)

            exceptions: List[Exception] = []
            for attempt_num, client in enumerate(
                self.model_selection_policy.get_clients()
            ):
                logger.info(
                    f"Starting attempt {attempt_num + 1} of {len(self.model_selection_policy)}"
                )

                try:
                    logger.debug(f"Received input: {input.escape()}")

                    conformed_input: Schema = input
                    if not self.InputSchema.is_empty():
                        conformed_input = input.conform(self.InputSchema, logger)
                        logger.debug(
                            f"Conformed input to {self.InputSchema}: {conformed_input.escape()}"
                        )

                    if not hasattr(self, "_prompt"):
                        raise ConfigurationError("Unit must define a prompt.")

                    prompt_message: PromptMessage = self.populate_prompt_message(
                        conformed_input, logger
                    )
                    logger.debug(f"Populated system prompt: {prompt_message.system}")
                    logger.debug(
                        f"Populated user prompt: {shorten_string(prompt_message.user)}"
                    )

                    in_tokens = len(client.encode(prompt_message.user))
                    out_tokens_estimate = np.mean(
                        self.shared.output_tokens or [0.0]
                    ).item()
                    ready = client.model.rate_limit.acquire(
                        {"requests": 1, "tokens": int(in_tokens + out_tokens_estimate)}
                    )
                    logger.debug(
                        f"Prepared in_tokens={in_tokens}, estimated out_tokens={out_tokens_estimate}"
                    )
                    if waiting := not ready.is_set():
                        logger.debug("Rate limit reached. Waiting...")
                    ready.wait()
                    if waiting:
                        logger.debug("Below rate limit again. Resuming...")

                    extractor: Extractor = (
                        self.extractor()
                        if isinstance(self.extractor, type)
                        else self.extractor
                    )
                    extractor.inject(unit=self)
                    logger.debug(f"Using extractor: {extractor}")

                    # exit early right before the inference call
                    if self.executor.is_complete.is_set():
                        logger.error(
                            "Exiting early since executor has been marked is_complete"
                        )
                        return  # type: ignore

                    try:
                        with DisableLogger("LiteLLM"):
                            response_stream, usage = extractor.extract(
                                client, prompt_message, logger
                            )
                            logger.info("Inference call succeeded")
                    except Exception as e:
                        logger.error(f"Inference call failed: {e}")
                        raise VerdictExecutionTimeError() from e

                    if isinstance(response_stream, Iterator):
                        logger.debug("Received streaming response")
                        for response in response_stream:
                            if (
                                self.should_stream_output
                                and self.streaming_layout_manager
                                and streaming_layout
                            ):
                                streaming_layout.update(response)
                    else:
                        response = response_stream
                    logger.debug(f"Received response: {response.escape()}")

                    out_tokens = usage.out_tokens
                    if usage.is_unknown() or usage.out_tokens == -1:
                        out_tokens = len(
                            client.encode(
                                str(response.model_dump())
                                if isinstance(response, Schema)
                                else response
                            )
                        )
                    self.shared.output_tokens.append(out_tokens)
                    client.model.rate_limit.release(
                        {"tokens": min(int(out_tokens - out_tokens_estimate), 0)}
                    )
                    logger.debug(f"Received out_tokens={out_tokens}")

                    try:
                        self.validate(conformed_input, response)
                    except Exception as e:
                        raise PostValidationError() from e
                    logger.debug("Unit.validate() successful")

                    try:
                        output = self.process(conformed_input, response)
                    except Exception as e:
                        raise PostProcessError() from e
                    logger.debug("Unit.process() successful")

                    try:
                        result = self._propagator(
                            self, Previous(self.dependencies), conformed_input, output
                        )  # type: ignore
                        logger.debug(f"Propagated result: {result}")
                        logger.info("Unit.execute() successful")
                        if call is not None:
                            call.set_outputs(result)
                        return result
                    except Exception as e:
                        raise PropagateError() from e
                except VerdictDeclarationTimeError as e:
                    logger.info("User-code/configuration error detected.")
                    raise e
                except Exception as e:
                    exceptions.append(e)
                    logger.info(f"Retrying after exception encountered: {e}")

            raise VerdictExecutionTimeError(
                f"Model Selection Policy {self.model_selection_policy} exhausted"
            ) from exceptions[-1]

    def validate(self, input: InputSchemaT, response: ResponseSchemaT) -> None:
        pass

    def process(
        self, input: InputSchemaT, response: ResponseSchemaT
    ) -> Union[OutputSchemaT, ResponseSchemaT]:
        return response


class LinkType:
    class How(Enum):
        @classmethod
        def from_str(cls, s: str) -> Self:
            return cls[s.upper()]

    class Inner(How):
        NONE = 0  #  U   U   U
        CHAIN = 1  #  U - U - U

    class Outer(How):
        DENSE = 0  #  U   U
        #  |\ /|
        #  |/ \|
        #  U   U

        BROADCAST = 1  #  U   U
        #  |   |
        #  |   |
        #  U   U

        CUMULATIVE = 2  #  U   U
        #  |\  |
        #  |  \|
        #  U   U

        LAST = 3  #  U   U
        #      |
        #      |
        #      U


class Layer(Graph[Node], Node, ModelConfigurable):  # type: ignore
    """
    Ordered list of units.
    """

    order: List[Node]

    how_inner: LinkType.Inner
    how_outer: LinkType.Outer

    root_idx: List[int]
    leaf_idx: List[int]

    @property
    def char(self) -> str:
        return f"Layer({self.how_inner}, {self.how_outer})"

    def __init__(
        self,
        nodes: Union[Node, List[Node]],
        repeat: int = 1,
        inner: Union[str, LinkType.Inner] = LinkType.Inner.NONE,
        outer: Union[str, LinkType.Outer] = LinkType.Outer.DENSE,
    ):
        Graph.__init__(self, node_type=(Node, List[Node]))  # type: ignore
        Node.__init__(self)

        if isinstance(nodes, Node):
            nodes = [nodes]

        self.order = []
        for i, _node in enumerate(cycle(nodes)):
            node = _node.copy()
            if hasattr(node, "idx"):
                node.idx(i + 1)

            node._ordering_timestamp = time.perf_counter()
            node.parent = self
            self.add(node)
            self.order.append(node)
            if len(self.order) >= len(nodes) * repeat:
                break

        self.how_inner = (
            LinkType.Inner.from_str(inner) if isinstance(inner, str) else inner
        )
        self.how_outer = (
            LinkType.Outer.from_str(outer) if isinstance(outer, str) else outer
        )

        self.root_idx = list(range(len(self.order)))
        self.leaf_idx = list(range(len(self.order)))

        if self.how_outer == LinkType.Outer.LAST:
            self.with_leaf(-1)

    def sort(self) -> List[Node]:
        return self.order

    def copy(self) -> Self:
        copy = type(self)(
            list(map(lambda node: node.copy(), self.order)),
            inner=self.how_inner,
            outer=self.how_outer,
        )
        copy.leaf_idx = self.leaf_idx
        copy.root_idx = self.root_idx
        return copy

    def clone(self) -> Self:
        clone = type(self)(
            list(map(lambda node: node.clone(), self.order)),
            inner=self.how_inner,
            outer=self.how_outer,
        )
        clone.leaf_idx = self.leaf_idx
        clone.root_idx = self.root_idx
        return clone

    def _materialize(
        self, context: MaterializationContext
    ) -> Tuple[Self, MaterializationContext]:
        if context.parent_branch:
            context.parent_branch = context.parent_branch.add(self.char)

        order = []
        context = context.copy()

        if self.how_inner == LinkType.Inner.CHAIN:
            with self.freeze_all_nodes():
                for i, node in enumerate(list(self.order)):
                    self.remove(node)

                    with node.freeze(), node.freeze_all_nodes():
                        self.add(
                            materialized := node._materialize(
                                context.copy().add_prefix(f"layer[{i}]")
                            )[0]
                        )

                    order.append(materialized)

                for previous, current in zip(order[:-1], order[1:]):
                    previous >> current

                self.order = order
                return self, context

        for i, node in enumerate(self.order):
            self.remove(node)
            self.add(
                materialized := node._materialize(
                    context.copy().add_prefix(f"layer[{i}]")
                )[0]
            )
            order.append(materialized)
        self.order = order
        return self, context

    def __rshift__(self, other: Union[Unit, "Layer", "Block"]) -> "Block":  # type: ignore
        return (Block() >> self) >> other

    def link(self, other: Union[Unit, "Layer", "Block"]) -> None:  # type: ignore
        if isinstance(other, Layer) and len(self.leaf_nodes) == len(other.root_nodes):
            if self.how_outer == LinkType.Outer.BROADCAST:
                assert len(self.leaf_nodes) == len(other.root_nodes), (
                    "Broadcast link requires equal number of leaf and root nodes"
                )
                with self.freeze(), other.freeze():
                    for a, b in zip(list(self.leaf_nodes), list(other.root_nodes)):
                        a.link(b)
                return
            elif self.how_outer == LinkType.Outer.CUMULATIVE:
                with (
                    self.freeze(),
                    other.freeze(),
                    self.freeze_leaf_nodes(),
                    other.freeze_root_nodes(),
                ):
                    for a_idx, a in enumerate(self.leaf_nodes):
                        for b in other.root_nodes[: a_idx + 1]:
                            a.link(b)
                return

        if isinstance(other, Unit):
            with self.freeze(), other.freeze():
                for node in self.leaf_nodes:
                    node.link(other)
            return

        with (
            self.freeze(),
            other.freeze(),
            self.freeze_leaf_nodes(),
            other.freeze_root_nodes(),
        ):
            for node in other.root_nodes:
                self.link(node)  # type: ignore

    def with_root(self, idx: Union[int, List[int]]) -> "Layer":
        self.root_idx = [idx] if isinstance(idx, int) else idx
        return self

    def with_leaf(self, idx: Union[int, List[int]]) -> "Layer":
        self.leaf_idx = [idx] if isinstance(idx, int) else idx
        return self

    def _idx_to_nodes(self, idx: List[int]) -> List[Node]:
        nodes = operator.itemgetter(*idx)(self.order)
        if not isinstance(nodes, tuple):
            nodes = (nodes,)
        return list(nodes)

    @property
    def root_nodes(self) -> List[Node]:
        return self._idx_to_nodes(self.root_idx)

    @property
    def leaf_nodes(self) -> List[Node]:
        return self._idx_to_nodes(self.leaf_idx)


class Block(Graph[Union[Unit, Layer]], Node, ModelConfigurable):
    use_root: bool = True

    @property
    def char(self) -> str:
        return "Block"

    def __init__(self, name: Optional[str] = None) -> None:
        Node.__init__(self, name)
        Graph.__init__(self, node_type=(Unit, Layer))  # type: ignore

        self.use_root = True

    def __rshift__(self, other: Union[Unit, Layer, "Block"]) -> "Block":  # type: ignore
        other.graph = self

        if isinstance(other, Block):  # merge blocks, update .graph
            with self.freeze(), other.freeze():
                for root in list(other.root_nodes):
                    if len(self.leaf_nodes) == 0:
                        self.add(root)
                    else:
                        for leaf in list(
                            self.root_nodes if self.use_root else self.leaf_nodes
                        ):
                            self.setup_link(leaf, root)
            self.nodes.update(other.nodes)
            return self

        if len(self.nodes) == 0:  # empty block just adds the node
            self.add(other)
            return self.leaf_view()

        with self.freeze():
            for leaf in list(self.root_nodes if self.use_root else self.leaf_nodes):
                self.setup_link(leaf, other)

        return self.leaf_view()

    def leaf_view(self) -> "Block":
        block = self.view()
        block.use_root = False
        return block

    def root_view(self) -> "Block":
        block = self.view()
        block.use_root = True
        return block

    def link(self, other: Union[Unit, Graph["Unit"]]) -> None:  # type: ignore
        with self.freeze():
            for node in list(self.leaf_nodes):
                node.link(other)  # type: ignore

    def materialize(self, _context: Optional[MaterializationContext] = None) -> "Block":
        return self._materialize(_context)[0]

    def _materialize(
        self, _context: Optional[MaterializationContext] = None, init: bool = True
    ) -> Tuple["Block", MaterializationContext]:
        context: MaterializationContext = (_context or MaterializationContext()).copy()
        if init:
            return self.copy()._materialize(context, init=False)

        materialized = Block()
        context.root_block = context.root_block or materialized
        if context.parent_branch:
            context.parent_branch = context.parent_branch.add("Block")

        source_to_destination: Dict[Union[Unit, Layer], Union[Unit, Layer]] = {}

        def map_to_destination(
            node: Union[Unit, Layer], context: MaterializationContext
        ) -> Tuple[Union[Unit, Layer], MaterializationContext]:
            if node not in source_to_destination:
                source_to_destination[node], context = node._materialize(
                    context.copy().add_prefix("block")
                )

            return source_to_destination[node], context

        for source, dependents in list(
            map(lambda node: (node, list(node.dependents)), self.sort())
        ):
            materialized_source, source_context = map_to_destination(source, context)
            materialized_source.graph = materialized

            if len(dependents) == 0 and isinstance(source, Unit):
                materialized.add(materialized_source)
                continue

            for dependent in dependents:
                materialized_dependent, dependent_context = map_to_destination(
                    dependent, source_context
                )  # type: ignore
                materialized_dependent.graph = materialized

                materialized_source.link(materialized_dependent)

        return materialized, context



import string
from typing import List, Optional

from verdict.core.primitive import Unit
from verdict.scale import DiscreteScale, LikertScale, Scale
from verdict.schema import Field, Schema


class JudgeUnit(Unit):
    """
    Direct score judge.
    """

    _char: str = "DirectScoreJudge"

    class ResponseSchema(Schema):
        score: Scale = LikertScale()

    def __init__(
        self, scale: Optional[Scale] = None, explanation: bool = False, **kwargs
    ):
        if scale is not None:
            self.OutputSchema = self.ResponseSchema = Schema.infer(score=scale)  # type: ignore

        self.scale = self.ResponseSchema.get_scale("score")

        if explanation:
            self.OutputSchema = self.ResponseSchema = self.ResponseSchema.prepend(
                explanation=str
            )  # type: ignore

        super().__init__(**kwargs)


class BestOfKJudgeUnit(Unit):
    _char: str = "BestOfKJudge"

    class InputSchema(Schema):
        # need a sentinel for empty, but using [''] for now makes type inference simple
        options: List[str] = Field(default=[""])

    class ResponseSchema(Schema):
        choice: DiscreteScale = DiscreteScale(["A", "B"])

    class OutputSchema(Schema):
        chosen: str

    def __init__(
        self,
        k: int = 2,
        response_options: Optional[DiscreteScale] = None,
        explanation: bool = False,
        original: bool = False,
        **kwargs,
    ):
        self.k = k
        self.explanation = explanation
        self.original = original

        response_options = response_options or DiscreteScale(
            list(string.ascii_uppercase[:k])
        )
        self.ResponseSchema = Schema.infer(choice=response_options)  # type: ignore

        self.scale = self.ResponseSchema.get_scale("choice")

        if explanation:
            self.ResponseSchema = self.ResponseSchema.prepend(explanation=str)  # type: ignore
            self.OutputSchema = self.OutputSchema.prepend(explanation=str)  # type: ignore

        super().__init__(**kwargs)

    def validate(self, input: InputSchema, response: ResponseSchema) -> bool:
        if self.original:
            assert len(input.options) != [""], (
                "Pass the options in InputSchema.options to index original choices."
            )
            assert len(input.options) == self.k, (
                f"Number of input options must equal {self.k}"
            )

    def process(self, input: InputSchema, response: ResponseSchema) -> OutputSchema:
        if not self.original:
            return response

        fields = {
            "chosen": input.options[
                self.scale.values.index(self.scale.value_mapping_fn(response.choice))
            ]
        }
        if self.explanation:
            fields["explanation"] = response.explanation

        return self.OutputSchema(**fields)


class PairwiseJudgeUnit(BestOfKJudgeUnit):
    """
    Judge unit for pairwise comparison (e.g., 'Response A' vs 'Response B').
    Options should be LLM responses.
    """

    _char: str = "PairwiseJudge"

    def __init__(
        self,
        response_options: Optional[DiscreteScale] = None,
        explanation: bool = False,
        original: bool = False,
        **kwargs,
    ):
        super().__init__(
            k=2,
            response_options=response_options,
            explanation=explanation,
            original=original,
            **kwargs,
        )


class CategoricalJudgeUnit(BestOfKJudgeUnit):
    """
    Judge unit for categorical decisions (e.g., 'Yes' or 'No', 'Harmful' or 'Not Harmful', 'Hallucination' or 'No Hallucination').
    Options should represent discrete categorical choices.
    """

    _char: str = "CategoricalJudge"

    def __init__(
        self,
        categories: Optional[DiscreteScale] = None,
        explanation: bool = False,
        original: bool = False,
        **kwargs,
    ):
        super().__init__(
            k=2,
            response_options=categories,
            explanation=explanation,
            original=original,
            **kwargs,
        )



"""
Rater agreement metrics
"""

from __future__ import annotations

import math
from dataclasses import dataclass
from typing import Callable, List, Optional, Union

import rich.box
import rich.console
import rich.layout
import rich.spinner
import rich.table


@dataclass
class Result:
    INVALID_RESULT_SENTINEL = float("nan")

    value: float
    percentage: bool = False

    def __format__(self, format_spec: str) -> str:
        if math.isnan(self.value):
            return "—"
        if "%" in format_spec:
            format_spec = format_spec.replace("%", "")
            return f"{self.value * 100:{format_spec}}%"
        return self.value.__format__(format_spec)

    def __str__(self) -> str:
        return f"{self:.2f%}" if self.percentage else f"{self:.2f}"

    def __rich__(self) -> str:
        return str(self)


spinner = rich.spinner.Spinner("simpleDotsScrolling")


def metric(
    percentage: bool = False,
) -> Callable[[Callable[..., Result]], Callable[..., Result]]:
    def decorator(func: Callable[..., Result]) -> Callable[..., Result]:
        def wrapper(*args, **kwargs) -> Result:
            df, ground_truth_col, prediction_col = args
            if prediction_col not in df.columns:
                return spinner

            try:
                return Result(value=func(*args, **kwargs), percentage=percentage)
            except Exception:
                return Result(
                    value=Result.INVALID_RESULT_SENTINEL, percentage=percentage
                )

        return wrapper

    return decorator


@metric(percentage=True)
def accuracy(df: "pd.DataFrame", ground_truth_col: str, prediction_col: str) -> Result:  # type: ignore
    return (df[ground_truth_col] == df[prediction_col]).mean()  # type: ignore


@metric()
def kappa(df: "pd.DataFrame", ground_truth_col: str, prediction_col: str) -> Result:  # type: ignore
    from sklearn.metrics import cohen_kappa_score  # type: ignore[import-untyped]

    return cohen_kappa_score(df[ground_truth_col], df[prediction_col])


@metric()
def kendall_tau(
    df: "pd.DataFrame", ground_truth_col: str, prediction_col: str
) -> Result:  # type: ignore
    from scipy.stats import kendalltau  # type: ignore[import-untyped]

    return kendalltau(df[ground_truth_col], df[prediction_col]).statistic


@metric()
def spearman_rho(
    df: "pd.DataFrame", ground_truth_col: str, prediction_col: str
) -> Result:  # type: ignore
    from scipy.stats import spearmanr  # type: ignore[import-untyped]

    return spearmanr(df[ground_truth_col], df[prediction_col]).statistic


@metric()
def krippendorff_alpha(
    df: "pd.DataFrame", ground_truth_col: str, prediction_col: str
) -> Result:  # type: ignore
    import pandas as pd

    data_interval = (
        pd.concat([df[ground_truth_col], df[prediction_col]], axis=1).to_numpy().T
    )

    import krippendorff  # type: ignore[import-untyped]

    return krippendorff.alpha(
        reliability_data=data_interval, level_of_measurement="interval"
    )


@dataclass
class ExperimentConfig:
    ground_truth_cols: List[str]
    prediction_cols: List[str]

    pivot_cols: Optional[List[str]] = None


def compute_stats_table(
    df: "pd.DataFrame", experiment_config: ExperimentConfig
) -> list:  # type: ignore
    """
    Compute the raw data for the stats table.

    Returns a list of rows, where each row is a tuple containing the data for a row in the table.
    """
    rows = []

    def _get_row(
        pivot_column: Optional[str],
        pivot_value: Optional[str],
        _df: "pd.DataFrame",
        ground_truth_col: str,
        prediction_col: str,
    ) -> tuple[str, str, str, str, str, Result, Result, Result, Result]:
        return (
            f"{pivot_column}={pivot_value}" if pivot_column and pivot_value else "—",
            str(len(_df)),
            ground_truth_col,
            prediction_col,
            "",
            accuracy(_df, ground_truth_col, prediction_col),
            kappa(_df, ground_truth_col, prediction_col),
            kendall_tau(_df, ground_truth_col, prediction_col),
            spearman_rho(_df, ground_truth_col, prediction_col),
        )

    for ground_truth_col, prediction_col in zip(
        experiment_config.ground_truth_cols, experiment_config.prediction_cols
    ):
        if experiment_config.pivot_cols:
            rows.append(_get_row("*", "*", df, ground_truth_col, prediction_col))
            for pivot_column in experiment_config.pivot_cols:
                for pivot_value, _df in df.groupby(pivot_column):
                    rows.append(
                        _get_row(
                            pivot_column,
                            str(pivot_value),
                            _df,
                            ground_truth_col,
                            prediction_col,
                        )
                    )
        else:
            rows.append(_get_row(None, None, df, ground_truth_col, prediction_col)[2:])

    return rows


def format_stats_table(
    rows: list, experiment_config: ExperimentConfig
) -> rich.table.Table:
    """
    Format the stats table using Rich from a list of rows.

    Returns a rich.table.Table object.
    """
    table = rich.table.Table(box=rich.box.HORIZONTALS, expand=True)

    if experiment_config.pivot_cols:
        table.add_column("Pivot", justify="left", no_wrap=True)
        table.add_column("N", justify="center", no_wrap=True)
    table.add_column("Ground Truth", justify="right", no_wrap=True)
    table.add_column("Prediction", justify="left", no_wrap=True)
    table.add_column()
    table.add_column("Acc.", justify="center")
    table.add_column("Cohen (κ)", justify="center")
    table.add_column("Kendall (τ)", justify="center")
    table.add_column("Spearman (ρ)", justify="center")

    for row in rows:
        table.add_row(*row)

    return table


def stats(
    df: "pd.DataFrame",
    experiment_config: ExperimentConfig,  # type: ignore
) -> tuple[rich.table.Table, list[tuple[Union[str, int]]]]:
    """
    Compute and format the stats table.

    Returns a rich.table.Table object.
    """
    rows = compute_stats_table(df, experiment_config)
    return format_stats_table(rows, experiment_config), rows


def get_experiment_layout(
    df: "pd.DataFrame", experiment_config: ExperimentConfig
) -> rich.layout.Layout:  # type: ignore
    len(
        df[experiment_config.pivot_cols].drop_duplicates()
    ) if experiment_config.pivot_cols else 1

    formatted_table, _ = stats(df, experiment_config)

    console = rich.console.Console()
    console = rich.console.Console(
        width=(console.size.width // 2) - 6,
    )
    with console.capture() as capture:
        console.print(formatted_table)

    height_requested = len(capture.get().split("\n"))

    return rich.layout.Layout(
        rich.panel.Panel(formatted_table, title=f"Agreement (N={len(df)} samples)"),
        size=height_requested,
        name="experiment",
    )


def display_stats(
    df: "pd.DataFrame",
    experiment_config: ExperimentConfig,  # type: ignore
) -> list[tuple[Union[str, int]]]:
    console = rich.console.Console()
    fmt_table, raw_table = stats(df, experiment_config)
    console.print(fmt_table)
    return raw_table



from __future__ import annotations

import threading
from collections import defaultdict
from types import SimpleNamespace
from typing import Any, Collection, Dict, List, Optional

import rich.layout
import rich.panel
import rich.text

from verdict.core.executor import ExecutionState
from verdict.schema import Schema


class StreamingLayoutManager:
    lock: threading.Lock

    streaming_panel: rich.panel.Panel
    layouts: Collection[rich.layout.Layout]

    class StreamingLayout:
        manager: "StreamingLayoutManager"
        unit: "Unit"  # noqa: F821 # type: ignore[name-defined]
        order: int

        text: rich.text.Text
        panel: rich.panel.Panel
        layout: rich.layout.Layout

        def __init__(
            self,
            manager: "StreamingLayoutManager",
            title: str,
            unit: "Unit",
            order: int,
        ):  # noqa: F821 # type: ignore[name-defined]
            self.manager = manager
            self.unit = unit
            self.order = order

            self.text = rich.text.Text("...")
            self.panel = rich.panel.Panel(
                rich.align.Align.left(
                    self.text,
                    vertical="top",
                ),
                title=title,
            )
            self.layout = rich.layout.Layout(self.panel)

        def update(self, response: Schema) -> None:
            self.panel = rich.panel.Panel(
                text := rich.text.Text(),
                title=self.panel.title,
                subtitle=self.panel.subtitle,
                border_style=self.panel.border_style,
                padding=self.panel.padding,
            )

            text.plain = ""
            for k, v in response.model_dump().items():
                text.append(f"{k}:", style="bold underline")
                text.append(f" {v}\n")

            self.text.plain = ""
            self.text.append(text)

            # detect overflow
            try:
                console = rich.console.Console()
                console = rich.console.Console(
                    width=(console.size.width // 2) - 6,
                )
                with console.capture() as capture:
                    console.print(self.panel)

                rendered_output = capture.get()
                height_requested = len(rendered_output.split("\n"))

                with self.manager.lock:
                    height_capacity = console.height // len(self.manager.children)
                    if height_requested > height_capacity:
                        height_overflow = height_requested - height_capacity
                        self.text.plain = ""
                        self.text.append(
                            "...\n"
                            + "\n".join(
                                map(
                                    lambda line: line.strip(" │"),
                                    rendered_output.split("\n")[
                                        2 + height_overflow : -3
                                    ],
                                )
                            )
                        )
            except Exception:
                self.text.plain = ""
                self.text.append(text)

    def __init__(self, layout: rich.layout.Layout, capacity: int = 5) -> None:
        self.lock = threading.Lock()

        self.layout = layout
        self.children: List[StreamingLayoutManager.StreamingLayout] = []

        self.capacity = capacity
        self.total = 0

    def add(self, unit: "Unit") -> StreamingLayout:  # noqa: F821 # type: ignore[name-defined]
        title = f"[{unit.char}]" + (
            f" [orange_red1]{unit.name}[/orange_red1]" if unit.name else ""
        )
        with self.lock:
            layout = StreamingLayoutManager.StreamingLayout(
                manager=self, title=title, unit=unit, order=self.total
            )

            # TODO: queue up running streamers, display as possible
            while len(self.children) >= self.capacity:
                # remove completed tasks first, then oldest tasks
                if not any(child.unit.completed for child in self.children):
                    return layout
                self.children.sort(
                    key=lambda child: (-int(child.unit.completed), child.order)
                )
                self.layout._children.remove(self.children.pop(0).layout)

            self.children.append(layout)
            self.layout.add_split(layout.layout)
            self.total += 1

            return layout


class BranchManager:
    spinner = {
        ExecutionState.WAITING_FOR_DEPENDENCIES: "bounce",
        ExecutionState.WAITING_FOR_RESOURCES: "dots11",
        ExecutionState.RUNNING: "dots3",
    }

    lock: threading._RLock
    branch: Optional[rich.tree.Tree]
    n_peers: int
    should_pin_output: bool

    visited_states: Dict[ExecutionState, int]

    def __init__(
        self,
        lock: threading.Lock,
        branch: Optional[rich.tree.Tree],
        n_peers: int,
        should_pin_output: bool,
    ) -> None:
        self.lock = threading.RLock()
        self.branch = branch
        self.n_peers = n_peers

        self.should_pin_output = should_pin_output
        self.length = 1 if self.should_pin_output else n_peers

        self.visited_states = defaultdict(int)

    def update(self, state: ExecutionState, unit: "Unit") -> None:  # noqa: F821 # type: ignore[name-defined]
        from verdict.core.primitive import Unit

        if not self.branch:
            return

        with self.lock:
            if self.visited_states[state] >= self.length:
                return
            self.visited_states[state] += 1

            text = f"[{unit.char}]" + (
                f" [orange_red1]{unit.name}[/orange_red1]" if unit.name else ""
            )
            if state == ExecutionState.WAITING_FOR_DEPENDENCIES:
                if self.n_peers <= 1:
                    table = rich.table.Table.grid(expand=True)
                    table.add_column(justify="left", ratio=3)  # status
                    table.add_column(
                        justify="right", ratio=1, no_wrap=True, overflow="ellipsis"
                    )  # result

                    groups = [
                        rich.spinner.Spinner(BranchManager.spinner[state], text=text)
                    ]
                    if type(unit).description is not Unit.description:
                        groups.append(f"\t{str(unit.description)}")  # type: ignore
                    table.add_row(rich.console.Group(*groups), "")
                    self.branch = self.branch.add(table)
                else:
                    # set progress bar
                    progress = rich.progress.Progress(
                        rich.progress.SpinnerColumn(
                            spinner_name=BranchManager.spinner[ExecutionState.RUNNING]
                        ),
                        rich.progress.BarColumn(bar_width=10),
                        rich.progress.MofNCompleteColumn(),  # TODO: running | completed/total
                        rich.progress.TextColumn(
                            "[progress.description]{task.description}"
                        ),
                        rich.progress.TimeElapsedColumn(),
                        rich.progress.TimeRemainingColumn(),
                    )
                    progress.add_task(text, total=self.length)
                    self.branch = self.branch.add(rich.console.Group(progress))
            elif state in [
                ExecutionState.WAITING_FOR_RESOURCES,
                ExecutionState.RUNNING,
            ]:
                if self.n_peers <= 1:
                    self.branch.label.columns[0]._cells[0].renderables[0] = (
                        rich.spinner.Spinner(  # type: ignore
                            BranchManager.spinner[state],
                            text=self.branch.label.columns[0]
                            ._cells[0]
                            .renderables[0]
                            .text,  # type: ignore
                        )
                    )
                # (pending_tasks): increment pending (just once!), use task.fields
            elif state == ExecutionState.COMPLETE:
                if self.n_peers <= 1:
                    self.branch.label.columns[0]._cells[0].renderables[0] = (
                        f"✔ {self.branch.label.columns[0]._cells[0].renderables[0].text}"  # type: ignore
                    )
                    if isinstance(unit.output, Schema):
                        unit_output = unit.output.model_dump()
                        if len(unit_output) == 1:
                            output: Any = unit_output[list(unit_output.keys())[0]]
                            if isinstance(output, float):
                                output = f"{output:.4f}"
                            output_str: str = str(output).split("\n")[0]
                        else:

                            def clean(output: Any) -> str:
                                if isinstance(output, float):
                                    return f"{output:.4f}"
                                elif isinstance(output, str) and len(output) > 4:
                                    return "…"
                                return str(output)

                            output_str = ", ".join(
                                [f"{k}: {clean(v)}" for k, v in unit_output.items()]
                            )

                        self.branch.label.columns[1]._cells[0] = rich.text.Text(
                            output_str, style="bright_black"
                        )  # type: ignore
                else:
                    self.branch.label.renderables[0].advance(task_id=0)  # type: ignore
                # (pending_tasks): decrement pending
            else:
                if self.n_peers <= 1:
                    self.branch.label.columns[0]._cells[0].renderables[0] = (
                        f"✘ {self.branch.label.columns[0]._cells[0].renderables[0].text}"  # type: ignore
                    )
                else:
                    column = next(
                        filter(
                            lambda c: isinstance(c, rich.progress.SpinnerColumn),
                            self.branch.label.renderables[0].columns,
                        )
                    )  # type: ignore
                    column.spinner = SimpleNamespace(render=lambda task: "✘")



import threading
import time
from abc import ABC, abstractmethod
from collections import deque
from enum import Enum
from queue import Queue
from types import SimpleNamespace
from typing import Dict, List, Optional, Protocol, Tuple, Union

from verdict.util.exceptions import VerdictSystemError


def disable() -> None:
    from verdict.config import state
    from verdict.util.log import logger

    state.rate_limiter_disabled = True
    logger.info(
        "Rate limiting is disabled. All requests will follow an UnlimitedRateLimiter."
    )


def enable() -> None:
    from verdict.config import state
    from verdict.util.log import logger

    state.rate_limiter_disabled = False
    logger.info(
        "Rate limiting is enabled. All requests will fallback to their configured RateLimitPolicy."
    )


class RateLimiterMetric(Enum):
    REQUESTS = "requests"
    TOKENS = "tokens"


class RateLimiter(ABC):
    @abstractmethod
    def acquire(self, value: Optional[int] = None) -> threading.Event:
        pass

    @abstractmethod
    def release(self, value: Optional[int] = None) -> None:
        pass

    @abstractmethod
    def copy(self) -> "RateLimiter":
        pass

    def __str__(self) -> str:
        return f"{self.__class__.__name__}()"


class UnlimitedRateLimiter(RateLimiter):
    def acquire(self, value: Optional[int] = None) -> threading.Event:
        event = threading.Event()
        event.set()
        return event

    def release(self, value: Optional[int] = None) -> None:
        pass

    def copy(self) -> "UnlimitedRateLimiter":
        return UnlimitedRateLimiter()


class ConcurrentRateLimiter(RateLimiter):
    # config
    max_concurrent: int

    # state
    waiting: Queue[threading.Event]
    lock: threading.RLock
    running: int

    def __init__(self, max_concurrent: int) -> None:
        self.waiting = Queue()

        self.max_concurrent = max_concurrent

        self.lock = threading.RLock()
        self.running = 0

    def copy(self) -> "ConcurrentRateLimiter":
        return ConcurrentRateLimiter(self.max_concurrent)

    def expire(self) -> None:
        with self.lock:
            while self.running < self.max_concurrent and not self.waiting.empty():
                event = self.waiting.get()
                self.running += 1
                event.set()

    def acquire(self, value: Optional[int] = None) -> threading.Event:
        event = threading.Event()
        self.waiting.put(event)

        with self.lock:
            self.expire()

        return event

    def release(self, value: Optional[int] = None) -> None:
        with self.lock:
            self.running -= 1
            self.expire()

    def __str__(self) -> str:
        return f"{self.__class__.__name__}(max={self.max_concurrent})"


class TimeWindowRateLimiter(RateLimiter):
    def __init__(
        self, max_value: int, window_seconds: int, smoothing_factor: float = 0.9
    ) -> None:
        self.max_value = max_value
        self.window_seconds = window_seconds
        self.smoothing_factor = smoothing_factor

        self.lock = threading.RLock()
        self.values: deque[Tuple[int, float]] = deque()
        self.waiting: deque[Tuple[threading.Event, int]] = deque()
        self._stop_event = threading.Event()

        # expiration thread
        self._expiration_thread = threading.Thread(
            target=self._expire_and_process, daemon=True
        )
        self._expiration_thread.start()

    def copy(self) -> "TimeWindowRateLimiter":
        return TimeWindowRateLimiter(
            self.max_value, self.window_seconds, self.smoothing_factor
        )

    def _expire_and_process(self) -> None:
        while not self._stop_event.is_set():
            with self.lock:
                self.expire()
                self._process_waiting_tasks()
            time.sleep(0.1)

    def expire(self) -> None:
        now = time.perf_counter()
        while self.values and now - self.values[0][1] > self.window_seconds:
            self.values.popleft()

    def current_sum(self) -> int:
        return sum(value for value, _ in self.values)

    def acquire(self, value: Optional[int] = None) -> threading.Event:
        if value is None:
            raise VerdictSystemError("Value is required for TimeWindowRateLimiter")

        event = threading.Event()
        with self.lock:
            self.expire()

            if self.current_sum() + value <= (self.max_value * self.smoothing_factor):
                self.values.append((value, time.perf_counter()))
                event.set()
            else:
                self.waiting.append((event, value))

        return event

    def _process_waiting_tasks(self) -> None:
        while self.waiting:
            wait_event, wait_value = self.waiting[0]
            if self.current_sum() + wait_value <= (
                self.max_value * self.smoothing_factor
            ):
                self.values.append((wait_value, time.perf_counter()))
                wait_event.set()
                self.waiting.popleft()
            else:
                break

    def release(self, value: Optional[int] = None) -> None:
        if value is None:
            raise VerdictSystemError("Value is required for TimeWindowRateLimiter")

        with self.lock:
            self.values.append((value, time.perf_counter()))
            self._process_waiting_tasks()

    def shutdown(self) -> None:
        """Stop the expiration thread."""
        self._stop_event.set()
        self._expiration_thread.join()

    def __del__(self) -> None:
        self.shutdown()

    def __str__(self) -> str:
        return f"{self.__class__.__name__}(max_value={self.max_value}, window_seconds={self.window_seconds}, smoothing_factor={self.smoothing_factor})"


RateLimitConfig = Dict[RateLimiter, Union[str, RateLimiterMetric]]


class MultiEvent(Protocol):
    def wait(self) -> None: ...

    def is_set(self) -> bool: ...


class RateLimitPolicy:
    # config
    rate_limiters: Dict[RateLimiter, RateLimiterMetric]

    def __init__(self, rate_limiters: RateLimitConfig) -> None:
        self.rate_limiters = {
            rate_limiter: RateLimiterMetric(metric)
            for rate_limiter, metric in rate_limiters.items()
        }

    def copy(self) -> "RateLimitPolicy":
        return RateLimitPolicy(
            {
                rate_limiter.copy(): metric
                for rate_limiter, metric in self.rate_limiters.items()
            }
        )

    def acquire(self, values: Dict[str, int] = {}) -> MultiEvent:
        events: List[threading.Event] = []
        for rate_limiter, metric in self.rate_limiters.items():
            events.append(rate_limiter.acquire(values.get(metric.value, 0)))

        return SimpleNamespace(
            wait=lambda: [event.wait() for event in events],
            is_set=lambda: all(event.is_set() for event in events),
        )

    def release(self, values: Dict[str, int] = {}) -> None:
        for rate_limiter, metric in self.rate_limiters.items():
            rate_limiter.release(values.get(metric.value, 0))

    def __str__(self) -> str:
        return f"{self.__class__.__name__}({' | '.join(f'{metric.value}@{rate_limiter}' for rate_limiter, metric in self.rate_limiters.items())})"

    def __repr__(self) -> str:
        return self.__str__()

    @staticmethod
    def of(rpm: int, tpm: int) -> "RateLimitPolicy":
        return RateLimitPolicy(
            {
                TimeWindowRateLimiter(max_value=rpm, window_seconds=60): "requests",
                TimeWindowRateLimiter(max_value=tpm, window_seconds=60): "tokens",
            }
        )

    @staticmethod
    def using(
        requests: Optional[Union[RateLimiter, List[RateLimiter]]] = [],
        tokens: Optional[Union[RateLimiter, List[RateLimiter]]] = [],
    ) -> "RateLimitPolicy":
        requests = requests if isinstance(requests, list) else [requests]
        tokens = tokens if isinstance(tokens, list) else [tokens]

        return RateLimitPolicy(
            {
                **{request: RateLimiterMetric.REQUESTS for request in requests},
                **{token: RateLimiterMetric.TOKENS for token in tokens},
            }
        )



import math
import random
import re
from abc import ABC, abstractmethod
from collections import defaultdict
from dataclasses import dataclass
from typing import Any, Dict, Iterator, List, Optional, Tuple, Type, Union

from loguru._logger import Logger
from pydantic import BaseModel
from typing_extensions import Self

from verdict import config
from verdict.model import ClientWrapper, Model, ModelSelectionPolicy
from verdict.prompt import PromptMessage
from verdict.scale import DiscreteScale
from verdict.schema import Schema
from verdict.util.exceptions import ConfigurationError, VerdictExecutionTimeError


@dataclass
class Usage:
    in_tokens: int
    out_tokens: int

    @staticmethod
    def unknown() -> "Usage":
        return Usage(in_tokens=-1, out_tokens=-1)

    def is_unknown(self) -> bool:
        return self.in_tokens == -1 and self.out_tokens == -1


class Extractor(ABC):
    """
    Represents a method of extracting a ResponseSchema from a provider call.

    Some examples:
        1. function-calling / structured output via `instructor`
        2. obtaining probability using logprobs on some token support (eg, `yes`, `no`)
        3. having a second LLM extract from a raw response string
    """

    response_schema: Type[Schema]
    streaming: bool = False

    @abstractmethod
    def extract(
        self,
        client_wrapper: ClientWrapper,
        prompt_message: PromptMessage,
        logger: Logger,
    ) -> Tuple[Union[Schema, Iterator[Schema]], Usage]:
        pass

    def inject(self, unit) -> None:
        self.response_schema = unit.ResponseSchema
        if hasattr(unit, "scale"):
            self.scale = unit.scale
        self.streaming = getattr(unit, "should_stream_output", False)

    def __str__(self) -> str:
        return f"{self.__class__.__name__}(streaming={self.streaming}, response_schema={self.response_schema.model_fields})"

    def __repr__(self) -> str:
        return self.__str__()

    @classmethod
    def format(cls) -> str:
        return f"{cls.__name__.replace('Extractor', '')}({{model_name}})"


class StructuredOutputExtractor(Extractor):
    def extract(
        self,
        client_wrapper: ClientWrapper,
        prompt_message: PromptMessage,
        logger: Logger,
    ) -> Tuple[Union[Schema, Iterator[Schema]], Usage]:
        assert getattr(self, "response_schema") is not None, (
            "StructuredOutputExtractor.response_schema must be set before calling extract()"
        )
        response = client_wrapper.function_calling_client(
            logger=logger,
            messages=(
                messages := prompt_message.to_messages(
                    add_nonce=client_wrapper.model.use_nonce
                )
            ),
            response_model=self.response_schema,
            streaming=self.streaming,
        )

        # TODO: add support / ping LiteLLM for image token usage tracking
        in_tokens = 0
        for message in messages:
            for content in message["content"]:
                if isinstance(content, str):
                    continue
                if content["type"] == "text":
                    in_tokens += len(client_wrapper.encode(content["text"]))

        usage = Usage(
            in_tokens=in_tokens,
            out_tokens=len(client_wrapper.encode(str(response.model_dump())))
            if not self.streaming
            else -1,
        )

        return response, usage


class RawExtractor(Extractor):
    field_name: str

    def inject(self, unit) -> None:
        super().inject(unit)

        model_fields = self.response_schema.model_fields
        assert len(model_fields) == 1, (
            "RawExtractor only supports a single output field"
        )

        [(self.field_name, field_info)] = model_fields.items()
        assert field_info.annotation is str, (
            "RawExtractor only supports a single `str` output field"
        )

    def extract(
        self,
        client_wrapper: ClientWrapper,
        prompt_message: PromptMessage,
        logger: Logger,
    ) -> Tuple[Union[Schema, Iterator[Schema]], Usage]:
        output = client_wrapper.raw_client(
            logger=logger,
            messages=(
                messages := prompt_message.to_messages(
                    add_nonce=client_wrapper.model.use_nonce
                )
            ),
            streaming=self.streaming,
        )

        def streaming_extract(output, messages) -> Iterator[Schema]:
            from litellm import stream_chunk_builder  # type: ignore[import-untyped]

            chunks = []
            for chunk in output:
                chunks.append(chunk)
                yield self.response_schema(
                    **{
                        self.field_name: stream_chunk_builder(chunks, messages=messages)
                        .choices[0]
                        .message.content
                    }
                )  # type: ignore

        if isinstance(output, Iterator):
            return streaming_extract(output, messages), Usage.unknown()
        else:
            usage = Usage(
                in_tokens=output.usage.prompt_tokens,
                out_tokens=output.usage.completion_tokens,
            )

            return self.response_schema(
                **{self.field_name: output.choices[0].message.content}
            ), usage


class CustomExtractor(RawExtractor, ABC):
    def inject(self, unit) -> None:
        Extractor.inject(self, unit)
        self.streaming = False

        self.original_response_schema = self.response_schema

        self.field_name = "output"
        self.response_schema = Schema.inline(**{self.field_name: str})

    @abstractmethod
    def post_extract(self, output: str, logger: Logger) -> Dict[str, Any]:
        pass

    def extract(
        self,
        client_wrapper: ClientWrapper,
        prompt_message: PromptMessage,
        logger: Logger,
    ) -> Tuple[Schema, Usage]:
        output, usage = super().extract(client_wrapper, prompt_message, logger)
        logger.debug(
            f"CustomExtractor {self.__class__.__name__} received output: {output.escape()}"
        )

        extracted = self.post_extract(output.output, logger)  # type: ignore
        logger.debug(
            f"CustomExtractor {self.__class__.__name__} completed post_extract: {extracted}"
        )
        if missing := set(self.original_response_schema.model_fields.keys()) - set(
            extracted.keys()
        ):
            raise ConfigurationError(
                f"Missing field(s) {missing} in CustomExtractor {self.__class__.__name__}."
            )

        try:
            return self.original_response_schema(**extracted), usage
        except Exception as e:
            raise ConfigurationError(
                f"CustomExtractor {self.__class__.__name__}.extract() failed to cast match '{extracted}' for field '{self.field_name}' in the output to {self.original_response_schema}. "
            ) from e


class RegexExtractor(CustomExtractor):
    fields: Dict[str, re.Pattern]

    FIRST_INT = r"[+-]?\d+"
    FIRST_FLOAT = r"[+-]?\d+(\.\d+)?"

    def __init__(self, fields: Dict[str, str]) -> None:
        self.fields = {field: re.compile(pattern) for field, pattern in fields.items()}

    def post_extract(self, output: str, logger: Logger) -> Dict[str, Any]:
        matches: Dict[str, Any] = {}
        for field in self.original_response_schema.model_fields.keys():
            match = (pattern := self.fields[field]).search(output)
            if match:
                match_str = match.group()
                logger.debug(
                    f"Found match {match_str} for field '{field}' with pattern '{pattern}'"
                )
                _type: Optional[Type] = self.original_response_schema.model_fields[
                    field
                ].annotation
                if _type is None:
                    raise ConfigurationError(
                        f"RegexExtractor.extract() failed to find a type for field '{field}'. Check ResponseSchema."
                    )
                logger.debug(f"Found type {_type} for field '{field}'")

                try:
                    matches[field] = _type(match_str)
                    logger.debug(f"Casted {match_str} to {_type}")
                except Exception as e:
                    raise VerdictExecutionTimeError(
                        f"RegexExtractor.extract() failed to cast match '{match_str}' for field '{field}' with pattern '{pattern}' in the output to {_type}. "
                    ) from e
            else:
                raise VerdictExecutionTimeError(
                    f"RegexExtractor.extract() failed to find a match for field '{field}' with pattern '{pattern}' in the output"
                )

        return matches

    def __str__(self) -> str:
        return f"{self.__class__.__name__}(streaming={self.streaming}, response_schema={self.response_schema.model_fields}, fields={self.fields})"


class PostHocExtractor(StructuredOutputExtractor):
    model_selection_policy: Optional[ModelSelectionPolicy] = None
    extract_client_wrappers: Optional[List[ClientWrapper]] = None

    def __init__(
        self,
        policy_or_name: Optional[Union[str, Model, List[Union[str, Model]]]] = None,
        retries: int = 1,
        **inference_parameters,
    ) -> None:
        if policy_or_name is not None:
            self.model_selection_policy = ModelSelectionPolicy.from_any(
                policy_or_name, retries, **inference_parameters
            )
            self.extract_client_wrappers = list(
                self.model_selection_policy.get_clients()
            )

        # NOTE: could make this portion stream too, but this makes the .execute code more complex
        self.raw_extractor = RawExtractor()
        super().__init__()

    def extract(
        self,
        client_wrapper: ClientWrapper,
        prompt_message: PromptMessage,
        logger: Logger,
    ) -> Tuple[Union[Schema, Iterator[Schema]], Usage]:
        self.raw_extractor.field_name = "output"
        self.raw_extractor.response_schema = Schema.inline(output=str)
        raw_response, raw_usage = self.raw_extractor.extract(
            client_wrapper, prompt_message, logger
        )

        if self.extract_client_wrappers is not None:
            logger.debug("Using custom post-hoc clients for extraction")
            extract_client_wrappers = self.extract_client_wrappers
        else:
            logger.debug("Using same client for extraction")
            extract_client_wrappers = [client_wrapper]

        for attempt_num, extract_client_wrapper in enumerate(extract_client_wrappers):
            logger.info(
                f"Starting attempt {attempt_num + 1} of {len(extract_client_wrappers)}"
            )

            ready = extract_client_wrapper.model.rate_limit.acquire(
                {
                    "requests": 1,
                    "tokens": raw_usage.in_tokens
                    + 20,  # estimate of the additional prompt below
                }
            )
            ready.wait()
            logger.debug(f"Acquired rate limit for {extract_client_wrapper.model}")

            try:
                response, usage = super().extract(
                    extract_client_wrapper,
                    PromptMessage(
                        system="You are an expert at extracting structured data from raw text. You faithfully use ONLY the Schema function/tool.",
                        user=f"""
Extract the following raw response into a structured format.

Raw response:
{raw_response.output}""",
                    ),
                    logger,
                )  # type: ignore
                logger.info("PostHoc inference call succeeded")
            except Exception as e:
                logger.error(f"PostHoc inference call failed: {e}. Retrying...")
                continue
            finally:
                extract_client_wrapper.model.rate_limit.release(
                    {"tokens": usage.out_tokens}
                )

            logger.debug(f"Received response: {response.escape()}")
            return response, raw_usage

    def format(self) -> str:
        if self.model_selection_policy is None:
            return f"{self.__class__.__name__.replace('Extractor', '')}({{model_name}} -> {{model_name}})"
        return f"{self.__class__.__name__.replace('Extractor', '')}({{model_name}} -> {self.model_selection_policy.char})"


class TokenProbabilityExtractor(Extractor):
    scale: DiscreteScale
    field_name: str

    def inject(self, unit) -> None:
        super().inject(unit)

        if not (hasattr(self.scale, "values") and hasattr(self.scale, "token_support")):
            raise ConfigurationError(
                "TokenProbabilityExtractor requires a Scale with a discrete list of values and a token_support() method"
            )

        model_fields = unit.ResponseSchema.model_fields
        assert len(model_fields) == 1, (
            f"{self.__class__.__name__} only supports a single output field"
        )
        [(self.field_name, field_info)] = model_fields.items()

        self.response_schema = Schema.inline(**{self.field_name: self.scale.T})

    # NOTE: technically this can stream, but it's useless in the single-token case
    def stream(self, stream: bool = False) -> Self:
        if stream:
            raise ConfigurationError(
                "TokenProbabilityExtractor does not support streaming"
            )
        return self

    def extract(
        self,
        client_wrapper: ClientWrapper,
        prompt_message: PromptMessage,
        logger: Logger,
    ) -> Tuple[Schema, Usage]:
        assert self.scale is not None, (
            "TokenProbabilityExtractor.scale must be set before calling extract()"
        )
        logger.debug(
            f"TokenProbabilityExtractor.extract with scale: {self.scale} (token_support: {self.scale.token_support()})"
        )

        messages = prompt_message.to_messages(client_wrapper.model.use_nonce)
        messages[-1]["content"] = config.TOKEN_EXTRACTOR_SPECIFICATION_PROMPT.format(
            content=messages[-1]["content"], scale_prompt=self.scale.prompt()
        )

        # TODO: support multi-tokens
        with client_wrapper.raw_client.defaults(
            temperature=0.0, logprobs=True, max_tokens=6, top_logprobs=20
        ):
            # logit_bias={client_wrapper.encode(token)[0]: 10 for token in self.scale.token_support()},
            response = client_wrapper.raw_client(
                logger=logger, messages=messages, streaming=self.streaming
            )

        logger.debug(
            f"TokenProbabilityExtractor {self.__class__.__name__} received response: {response.choices[0].message.content}"
        )

        usage = Usage(
            in_tokens=response.usage.prompt_tokens,
            out_tokens=response.usage.completion_tokens,
        )

        distribution: Dict[Any, float] = defaultdict(float)
        if hasattr(choice := response.choices[0], "logprobs"):
            logprobs = choice.logprobs
            logger.debug(
                f"TokenProbabilityExtractor {self.__class__.__name__} received logprobs: {logprobs}"
            )
            if isinstance(logprobs, dict):
                logprobs = logprobs["content"][0]["top_logprobs"]
            elif isinstance(logprobs, BaseModel):
                logprobs = logprobs.model_dump()
            else:
                raise ConfigurationError(f"Unsupported logprobs format: {logprobs}")
            if logprobs["content"] is not None:
                probabilities = {
                    lp["token"]: math.exp(lp["logprob"])
                    for lp in logprobs["content"][0]["top_logprobs"]
                    if lp["token"] in self.scale.token_support()
                }
            else:  # TogetherAI
                probabilities = {
                    token: math.exp(lp)
                    for token, lp in zip(logprobs["tokens"], logprobs["token_logprobs"])
                    if token in self.scale.token_support()
                }

            norm = sum(probabilities.values()) if len(probabilities) > 0 else 1
            for token in self.scale.token_support():
                distribution[self.scale.value_mapping_fn(token)] += (
                    probabilities.get(token, 0) / norm
                )
        else:
            raise ConfigurationError(f"""{client_wrapper.model} does not return logprobs.
Switch to a different model provider or to a non-TokenProbabilityExtractor.""")

        logger.debug(
            f"TokenProbabilityExtractor {self.__class__.__name__} received distribution over token support: {distribution}"
        )
        return Schema.inline(distribution=Dict[self.scale.T, float])(
            distribution=distribution
        ), usage  # type: ignore

    def __str__(self) -> str:
        return (
            f"{self.__class__.__name__}(streaming={self.streaming}, scale={self.scale})"
        )


class ArgmaxScoreExtractor(TokenProbabilityExtractor):
    def extract(
        self,
        client_wrapper: ClientWrapper,
        prompt_message: PromptMessage,
        logger: Logger,
    ) -> Tuple[Schema, Usage]:
        response, usage = super().extract(client_wrapper, prompt_message, logger)
        token_probability_distribution = response.distribution  # type: ignore[attr-defined]

        max_token, _ = max(token_probability_distribution.items(), key=lambda x: x[1])
        return self.response_schema(**{self.field_name: max_token}), usage


class SampleScoreExtractor(TokenProbabilityExtractor):
    def extract(
        self,
        client_wrapper: ClientWrapper,
        prompt_message: PromptMessage,
        logger: Logger,
    ) -> Tuple[Schema, Usage]:
        response, usage = super().extract(client_wrapper, prompt_message, logger)
        token_probability_distribution = response.distribution  # type: ignore[attr-defined]

        token, _ = random.choices(
            list(token_probability_distribution.items()),
            weights=list(token_probability_distribution.values()),
        )[0]
        return self.response_schema(**{self.field_name: token}), usage


class WeightedSummedScoreExtractor(TokenProbabilityExtractor):
    def inject(self, unit) -> None:
        super().inject(unit)

        if self.scale.T not in (bool, int, float):
            raise ConfigurationError(
                "WeightedSummedScoreExtractor requires a Scale with boolean, integer, or float values"
            )

        # special case where the output is a float regardless of using a DiscreteScale
        self.response_schema = Schema.inline(**{self.field_name: float})

    def extract(
        self,
        client_wrapper: ClientWrapper,
        prompt_message: PromptMessage,
        logger: Logger,
    ) -> Tuple[Schema, Usage]:
        response, usage = super().extract(client_wrapper, prompt_message, logger)
        token_probability_distribution = response.distribution  # type: ignore[attr-defined]

        weighted = Schema.of(
            output=sum(
                token * probability
                for token, probability in token_probability_distribution.items()
            )
        )
        return self.response_schema(**{self.field_name: weighted.output}), usage  # type: ignore



import verdict.util.log  # noqa: F401 type: ignore[unused-import]
from verdict.core.pipeline import Pipeline
from verdict.core.primitive import Block, Layer, Unit
from verdict.transform import __all__ as transform_all

__all__ = ["Unit", "Layer", "Block", "Pipeline", *transform_all]



import copy
from dataclasses import dataclass
from itertools import cycle
from typing import List, Optional

from verdict.core.primitive import Unit
from verdict.schema import Field, Schema


@dataclass
class Message:
    role_name: str
    message: str

    def __init__(self, role_name: str, message: str) -> None:
        self.role_name = role_name
        self.message = message

    def __str__(self) -> str:
        return f"<{self.role_name}>\n{self.message}\n</{self.role_name}>"

    def __repr__(self) -> str:
        return str(self)


class Conversation:
    history: List[Message]

    def __init__(self, history: Optional[List[Message]] = None) -> None:
        self.history = history or [Message("<START>", "")]

    def get_roles(self) -> List[str]:
        all_roles = list(map(lambda m: m.role_name, self.history[1:]))

        roles = []
        seen = set()
        for role in all_roles:
            if role not in seen:
                roles.append(role)
                seen.add(role)
        return roles

    def with_roles(self, roles: List[str]) -> "Conversation":
        assert len(roles) == len(self.get_roles()), (
            "Number of roles must match existing role structure"
        )
        for message, role in zip(self.history[1:], cycle(roles)):
            message.role_name = role

        return self

    def with_context(self) -> "Conversation":
        other = self.copy()
        other._with_context = True
        return other

    def __str__(self) -> str:
        starting_index = 0 if getattr(self, "_with_context", False) else 1
        return "\n\n".join(map(str, self.history[starting_index:]))

    def __repr__(self) -> str:
        return str(self)

    def copy(self) -> "Conversation":
        return copy.deepcopy(self)


class ConversationalUnit(Unit):
    @property
    def _char(self) -> str:
        return self.role_name

    class InputSchema(Schema):
        conversation: Conversation = Field(default_factory=Conversation)

    class ResponseSchema(Schema):
        response: str

    class OutputSchema(Schema):
        conversation: Conversation
        response: str

    def __init__(self, role_name: str, number: bool = True, **kwargs) -> None:
        super().__init__(**kwargs)
        self.role_name = role_name
        self.number = number

    def idx(self, value: Optional[int] = None) -> int:
        if value is not None and self.number and "#" not in self.role_name:
            self.role_name = f"{self.role_name} #{value}"
        return super().idx(value)

    def process(self, input: InputSchema, response: ResponseSchema) -> OutputSchema:
        conversation = copy.deepcopy(input.conversation)
        conversation.history.append(
            Message(role_name=self.role_name, message=response.response)
        )
        return self.OutputSchema(conversation=conversation, response=response.response)



from __future__ import annotations

import copy
import threading
from dataclasses import dataclass
from typing import Any, List, Optional, Set

from verdict.core.executor import Task
from verdict.core.visualization import BranchManager
from verdict.schema import Schema
from verdict.util.exceptions import ConfigurationError


# fresh on init/copy, shared on clone
@dataclass
class SynchronizationState:
    lock: threading.Lock

    peers: Set["Task"]  # n_peers

    shared: bool
    branch_lock: threading.Lock
    branch: Optional[BranchManager]

    # shared_output logic
    candidate_assigned: bool
    shared_output: threading.Condition
    output: Optional[Schema]

    output_tokens: List[float]

    def __init__(self) -> None:
        self.lock = threading.Lock()

        self.peers = set()

        self.shared = False  # sets to True on clone
        self.branch_lock = threading.Lock()
        self.branch = None

        self.candidate_assigned = False
        self.shared_output = threading.Condition()
        self.output = None

        self.output_tokens = []

    def add(self, peer: "Unit") -> "SynchronizationState":  # noqa: F821 # type: ignore[name-defined]
        with self.lock:
            self.peers.add(peer)
            self.shared = True
        return self


class UserState:
    def __init__(self) -> None:
        object.__setattr__(self, "_dict", {})

    def __getattr__(self, key: str) -> Any:
        try:
            return self._dict[key]
        except KeyError:
            raise ConfigurationError(f"'UserState' object has no attribute '{key}'")

    def __setattr__(self, key: str, value: Any) -> None:
        self._dict[key] = value

    def copy(self) -> "UserState":
        new = UserState()
        object.__setattr__(
            new, "_dict", {k: copy.copy(v) for k, v in self._dict.items()}
        )
        return new



import ast
import inspect
import random
import re
import string
import textwrap
from abc import ABC, abstractmethod
from typing import Any, Dict, List, NamedTuple, Optional, Set, Tuple, Type, Union

from loguru._logger import Logger
from typing_extensions import Self

from verdict.image import Image
from verdict.schema import Schema
from verdict.util.exceptions import PromptError

SECTION_REGEX = re.compile(r"@(\w+)(.*?)(?=@\w+|$)", re.DOTALL)


class PromptRegistry(type):
    _registry: Dict[str, Type["Prompt"]] = {}

    @staticmethod
    def extract_sections(template: str) -> Tuple[Optional[str], Optional[str], bool]:
        matches = SECTION_REGEX.findall(template)

        sections = {f"@{section}": content for section, content in matches}

        # Assign leading text to '@user' only if it exists and '@user' is not explicitly defined
        leading_text = template.split("@", 1)[0]
        if leading_text and "@user" not in sections:
            sections["@user"] = leading_text

        return sections.get("@system"), sections.get("@user"), "@no_format" in sections

    @staticmethod
    def strip_prompt_template(prompt: str) -> str:
        lines = prompt.splitlines()

        # Remove leading empty lines
        while lines and not lines[0].strip():
            lines.pop(0)

        # Remove trailing empty lines
        while lines and not lines[-1].strip():
            lines.pop()

        # Unindent the remaining lines and reassemble the docstring
        return "\n".join(line.lstrip() for line in lines)

    def __new__(cls, name, bases, dct) -> Type["Prompt"]:
        klass = super().__new__(cls, name, bases, dct)

        # only add named classes to the registry
        if name and not name.startswith("_") and name != "Prompt":
            cls._registry[name] = klass

        if not klass.__doc__:
            raise PromptError(f"Prompt {klass.__name__} must have __doc__.")

        if "" in (klass.extract_keys(klass.__doc__)):
            raise PromptError(f"All keys must be named in Prompt {klass.__name__}.")

        klass.system_prompt_template, klass.user_prompt_template, no_format = (
            klass.extract_sections(klass.__doc__)
        )
        assert klass.user_prompt_template is not None, (
            "User prompt template must be specified."
        )

        if not no_format:
            klass.user_prompt_template = klass.strip_prompt_template(
                klass.user_prompt_template
            )
            if klass.system_prompt_template:
                klass.system_prompt_template = klass.strip_prompt_template(
                    klass.system_prompt_template
                )

        return klass

    def __getattr__(self, name: str) -> Type["Prompt"]:
        if name in self._registry:
            return self._registry[name]
        else:
            raise PromptError(f"Prompt {name} not found in PromptRegistry._registry.")

    @staticmethod
    def compatible_prompts(unit: Type["Promptable"]) -> List[Type["Prompt"]]:
        return [
            prompt
            for prompt in PromptRegistry._registry.values()
            if prompt().supports(unit)
        ]


class PromptMessage(NamedTuple):
    system: Optional[str]
    user: str
    input_schema: Optional[Schema] = None

    def _get_image_values(self, input_schema: Optional[Schema] = None) -> List:
        image_values = []
        if input_schema is None:
            return image_values

        for field_value in input_schema.model_dump().values():
            if "type" in field_value and "image/" in field_value["type"]:
                image_values.append(field_value)

        return image_values

    def to_messages(
        self, add_nonce: bool = False
    ) -> List[Dict[str, str | List[Dict[str, str | Dict[str, str]]]]]:
        nonce = (
            "".join(random.choices(string.ascii_letters, k=10)) + "\n"
            if add_nonce
            else ""
        )
        content = [{"type": "text", "text": nonce + self.user}]

        image_values = self._get_image_values(self.input_schema)
        if image_values:
            for image in image_values:
                content.append(
                    {
                        "type": "image_url",
                        "image_url": {
                            "url": f"data:{image['type']};base64,{image['data']}"
                        },
                    }
                )

        messages = [{"role": "user", "content": content}]
        if self.system:
            messages.insert(0, {"role": "system", "content": nonce + self.system})
        return messages


RESERVED_KEYS = set(["input", "unit", "previous", "source", "prompt"])


class Prompt(metaclass=PromptRegistry):
    """
    A prompt template that can be used to generate a prompt for an LLM.
    Supports formatting with the keys of a Schema / pydantic.BaseModel.

    {key} will be replaced with the value of the key in the input at runtime.
    Indentation is preserved in the output.
    """

    system_prompt_template: Optional[str]
    user_prompt_template: str

    caller_locals: Dict[str, Any] = {}

    @staticmethod
    def from_template(template: str) -> "Prompt":
        return type("", (Prompt,), {"__doc__": template})()

    @staticmethod
    def extract_keys(template: str, exclude_reserved: bool = True) -> Set[str]:
        def extract_variables(template: str) -> Set[str]:
            variables = set()

            placeholder_pattern = r"(?<!\{)\{([^{}]+?)\}(?!\})"
            placeholders = re.findall(placeholder_pattern, template)

            for placeholder in placeholders:
                try:
                    tree = ast.parse(placeholder, mode="eval")

                    class VariableVisitor(ast.NodeVisitor):
                        def visit_Name(self, node) -> None:
                            variables.add(node.id)

                        def visit_Attribute(self, node) -> None:
                            while isinstance(node, ast.Attribute):
                                node = node.value
                            if isinstance(node, ast.Name):
                                variables.add(node.id)

                    VariableVisitor().visit(tree)
                except SyntaxError:
                    variables.add(placeholder.split(".")[0])

            return variables

        if not exclude_reserved:
            return extract_variables(template)
        return extract_variables(template) - RESERVED_KEYS

    def get_all_keys(self) -> Set[str]:
        keys = set()

        if self.system_prompt_template:
            keys |= Prompt.extract_keys(
                self.system_prompt_template, exclude_reserved=False
            )

        keys |= Prompt.extract_keys(self.user_prompt_template, exclude_reserved=False)

        return keys

    @property
    def keys(self) -> Set[str]:
        return self.get_all_keys() - RESERVED_KEYS - set(self.caller_locals.keys())

    def format(
        self,
        input_schema: Schema,
        unit: "Promptable",
        previous: Any,
        source: Schema,
        logger: Optional[Logger] = None,
    ) -> PromptMessage:
        for illegal_symbol in ["{{", "}}"]:
            for prompt in [self.system_prompt_template, self.user_prompt_template]:
                if prompt is not None and illegal_symbol in prompt:
                    if logger:
                        logger.warning(
                            f"Prompt contains '{illegal_symbol}'. Variable likely not getting evaluated."
                        )

        format_kwargs = (
            {"input": input_schema}
            | {"unit": unit}
            | {"previous": previous}
            | {"source": source}
        )

        for key, value in self.caller_locals.items():
            if key not in format_kwargs and not key.startswith("__"):
                format_kwargs[key] = value

        return PromptMessage(
            system=(
                self.auto_format(self.system_prompt_template, format_kwargs)
                if self.system_prompt_template
                else None
            ),
            user=self.auto_format(self.user_prompt_template, format_kwargs),
            input_schema=input_schema,
        )

    @staticmethod
    def auto_format(template: str, context: Dict[str, Any]) -> str:
        single_placeholder_pattern = r"(?<!\{)\{([^{}]+?)\}(?!\})"
        matches = re.findall(single_placeholder_pattern, template)

        for match in matches:
            try:
                value = eval(match, {}, context)
                template = template.replace(f"{{{match}}}", str(value))
            except Exception as e:
                raise PromptError(
                    textwrap.dedent(
                        f"""
                        Failed to evaluate Prompt placeholder '{match}' in the following context.

                        Context: {context}
                        """
                    )
                ) from e

        return template


class Promptable(ABC):
    _prompt: Prompt

    def prompt(self, prompt: Union[str, Prompt]) -> Self:
        if isinstance(prompt, str):
            self._prompt = Prompt.from_template(prompt)
        else:
            self._prompt = prompt

        frame = inspect.currentframe().f_back
        self._prompt.caller_locals = frame.f_locals

        return self

    @abstractmethod
    def populate_prompt_message(self, input: Schema, logger: Logger) -> PromptMessage:
        pass



import statistics
from operator import attrgetter
from typing import Any, Callable, List, Optional, Union

from verdict.core.primitive import Unit
from verdict.schema import Schema
from verdict.util.exceptions import VerdictExecutionTimeError
from verdict.util.misc import lightweight
from verdict.util.tracing import ExecutionContext


@lightweight
class MapUnit(Unit):
    _char: str = "Map"

    accumulate: bool = True
    map_func: Callable[[Union[Any, List[Any]]], Union[Any, List[Any]]]

    def __init__(
        self,
        map_func: Callable[[Union[Any, List[Any]]], Union[Any, List[Any]]],
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.data.map_func = lambda x: map_func(x)

    class InputSchema(Schema):
        values: Union[Any, List[Any]]

    class ResponseSchema(Schema):
        values: Union[Any, List[Any]]

    def execute(
        self, input: InputSchema, execution_context: Optional[ExecutionContext] = None
    ) -> ResponseSchema:
        if execution_context is None:
            execution_context = ExecutionContext()

        call_name = (
            self.__class__.__name__
            or getattr(self, "_char", None)
            or getattr(self, "char", None)
        )

        with execution_context.trace_call(
            name=call_name,
            inputs={"input": input, "unit": self},
        ) as call:
            try:
                values = input.values
                if len(values) == 1:
                    values = values[0]

                output = self.data.map_func(values)
                if isinstance(output, Schema):
                    self.OutputSchema = self.ResponseSchema = output.__class__  # type: ignore
                    result = output  # type: ignore
                else:
                    result = MapUnit.ResponseSchema(values=output)

                if call is not None:
                    call.set_outputs(result)
                return result
            except Exception as e:
                raise VerdictExecutionTimeError(
                    f"Failed to execute MapUnit: {e} on {input}"
                ) from e


class FieldMapUnit(MapUnit):
    fields: List[str]

    def __init__(
        self,
        map_func: Callable[[Union[Any, List[Any]]], Union[Any, List[Any]]],
        fields: Union[str, List[str]],
        **kwargs,
    ):
        super().__init__(map_func, **kwargs)
        self.fields = fields if isinstance(fields, list) else [fields]

    def execute(
        self,
        input: MapUnit.InputSchema,
        execution_context: Optional[ExecutionContext] = None,
    ) -> MapUnit.ResponseSchema:
        if execution_context is None:
            execution_context = ExecutionContext()

        call_name = (
            self.__class__.__name__
            or getattr(self, "_char", None)
            or getattr(self, "char", None)
        )

        with execution_context.trace_call(
            name=call_name,
            inputs={"input": input, "unit": self},
        ) as call:
            values = input.values

            if len(self.fields) == 0 and len(values) > 0:
                self.fields = list(values[0].model_fields.keys())

            assert all(field in values[0].model_fields for field in self.fields), (
                f"Fields {self.fields} not a subset of input {input.values}"
            )

            result = Schema.of(
                **{  # type: ignore
                    field: self.data.map_func(list(map(attrgetter(field), values)))
                    for field in self.fields
                }
            )

            if call is not None:
                call.set_outputs(result)
            return result

    @staticmethod
    def from_fn(
        fn: Callable[[List[Any]], Any], name: str
    ) -> Callable[[Union[str, List[str]]], "FieldMapUnit"]:
        return lambda fields=[]: FieldMapUnit(fn, fields, name=name)


class MeanPoolUnit(FieldMapUnit):
    def __init__(self, fields: Union[str, List[str]] = []):
        super().__init__(statistics.mean, fields, name="MeanPool")


class MaxPoolUnit(FieldMapUnit):
    def __init__(self, fields: Union[str, List[str]] = []):
        super().__init__(statistics.mode, fields, name="MaxPool")


class MeanVariancePoolUnit(FieldMapUnit):
    def __init__(self, fields: Union[str, List[str]] = []):
        super().__init__(
            lambda x: Schema.inline(mean=float, variance=float)(  # type: ignore
                mean=statistics.mean(x),
                variance=statistics.variance(x) if len(x) > 1 else 0.0,
            ),
            fields,
            name="MeanVariancePool",
        )


__all__ = [
    "MapUnit",
    "MeanPoolUnit",
    "MeanVariancePoolUnit",
    "MaxPoolUnit",
]



import os
import sys
from typing import Optional

from loguru import logger as base_logger

from verdict import config

# logging
VERDICT_COMPACT_LOG_FORMAT: str = "<green>{time:HH:mm:ss}</green> | <level>{level: <8}</level> | <red>{extra[unit]: >80} T={extra[thread_id]: <5}</red> | <cyan>{function}</cyan> - <level>{message:.150}...</level>"
VERDICT_VERBOSE_LOG_FORMAT: str = "<green>{time:YYYY-MM-DD HH:mm:ss.SSS}</green> | <level>{level: <8}</level> | <red>{extra[unit]: >80} T={extra[thread_id]: <5}</red> | <cyan>{name}</cyan>:<cyan>{function}</cyan>:<cyan>{line}</cyan> - <level>{message}</level>"

logger = base_logger.bind(thread_id="main", unit="")
logger.remove()


def init_logger(name: Optional[str] = None) -> None:
    global logger
    logger.remove()
    logger.add(sys.stderr, format=VERDICT_COMPACT_LOG_FORMAT, level="CRITICAL")

    if config.DEBUG or (log_level := os.getenv("LOG_LEVEL", False)):
        logger.add(
            sys.stderr, format=VERDICT_COMPACT_LOG_FORMAT, level=log_level or "DEBUG"
        )

    if name and not os.getenv("VERDICT_NO_LOG", False):
        logger.add(
            config.VERDICT_LOG_DIR / f"{name}_{{time}}.log",
            format=VERDICT_VERBOSE_LOG_FORMAT,
        )
